"""
An interface to output the data in various format
"""

import numpy
import os
import sys

from pysph.base.particle_array import ParticleArray
from pysph.base.utils import get_particles_info, get_particle_array
from pysph import has_h5py

output_formats = ('hdf5', 'npz')
COMPRESSION_LEVEL = 6


def _to_str(s):
    if isinstance(s, bytes) and sys.version_info[0] > 2:
        return s.decode('utf-8')
    else:
        return str(s)


def gather_array_data(all_array_data, comm):
    """Given array_data from the current processor and an MPI
    communicator,return a joined array_data from all processors
    on rank 0 and the same array_data on the other machines.
    """

    array_names = all_array_data.keys()

    # gather the data from all processors
    collected_data = comm.gather(all_array_data, root=0)

    if comm.Get_rank() == 0:
        all_array_data = {}
        size = comm.Get_size()

        # concatenate the arrays
        for array_name in array_names:
            array_data = {}
            all_array_data[array_name] = array_data

            _props = collected_data[0][array_name].keys()
            for prop in _props:
                data = [collected_data[pid][array_name][prop]
                        for pid in range(size)]
                prop_arr = numpy.concatenate(data)
                array_data[prop] = prop_arr
    return all_array_data


class Output(object):
    """ Class that handles output for simulation """
    def __init__(self, detailed_output=False, only_real=True, mpi_comm=None,
                 compress=False):
        self.compress = compress
        self.detailed_output = detailed_output
        self.only_real = only_real
        self.mpi_comm = mpi_comm

    def dump(self, fname, particles, solver_data):
        self.particle_data = dict(get_particles_info(particles))
        self.all_array_data = {}
        for array in particles:
            self.all_array_data[array.name] = array.get_property_arrays(
                all=self.detailed_output,
                only_real=self.only_real
                )
        mpi_comm = self.mpi_comm
        if mpi_comm is not None:
            self.all_array_data = gather_array_data(
                self.all_array_data, mpi_comm
            )
        self.solver_data = solver_data
        if mpi_comm is None or mpi_comm.Get_rank() == 0:
            self._dump(fname)

    def load(self, fname):
        return self._load(fname)

    def _dump(self, fname):
        """ Implement the method for writing the output to a file here """
        raise NotImplementedError()

    def _load(self, fname):
        """ Implement the method for loading from file here """
        raise NotImplementedError()


def _dict_bytes_to_str(d):
    # This craziness is needed as if the npz file is saved in Python2
    # then all the strings are bytes and if this is loaded in Python 3,
    # the keys will be bytes and not strings leading to strange errors.
    res = {}
    for key, value in d.items():
        if isinstance(value, dict):
            value = _dict_bytes_to_str(value)
        if isinstance(value, bytes):
            value = _to_str(value)
        if isinstance(value, list):
            if value and isinstance(value[0], bytes):
                value = [_to_str(x) for x in value]
        res[_to_str(key)] = value
    return res


def _get_dict_from_arrays(arrays):
    arrays.shape = (1,)
    res = arrays[0]
    if res and isinstance(list(res.keys())[0], bytes):
        return _dict_bytes_to_str(res)
    else:
        return res


class NumpyOutput(Output):

    def _dump(self, filename):
        save_method = numpy.savez_compressed if self.compress else numpy.savez
        output_data = {"particles": self.particle_data,
                       "solver_data": self.solver_data}
        for name, arrays in self.all_array_data.items():
            self.particle_data[name]["arrays"] = arrays
        save_method(filename, version=2, **output_data)

    def _load(self, fname):
        data = numpy.load(fname, encoding='bytes', allow_pickle=True)

        if 'version' not in data.files:
            msg = "Wrong file type! No version number recorded."
            raise RuntimeError(msg)
        ret = {}
        ret["arrays"] = {}
        version = data['version']
        solver_data = _get_dict_from_arrays(data["solver_data"])
        ret["solver_data"] = solver_data

        if version == 1:
            arrays = _get_dict_from_arrays(data["arrays"])
            for array_name in arrays:
                array = get_particle_array(name=array_name,
                                           **arrays[array_name])
                ret["arrays"][array_name] = array

        elif version == 2:
            particles = _get_dict_from_arrays(data["particles"])

            for array_name, array_info in particles.items():
                for prop, data in array_info['arrays'].items():
                    array_info['properties'][prop]['data'] = data
                array = ParticleArray(name=array_name,
                                      constants=array_info["constants"],
                                      **array_info["properties"])
                array.set_output_arrays(
                    array_info.get('output_property_arrays', [])
                )
                ret["arrays"][array_name] = array

        else:
            raise RuntimeError("Version not understood!")
        return ret


class HDFOutput(Output):

    def _dump(self, filename):
        import h5py
        with h5py.File(filename, 'w') as f:
            solver_grp = f.create_group('solver_data')
            particles_grp = f.create_group('particles')
            for ptype, pdata in self.particle_data.items():
                ptype_grp = particles_grp.create_group(ptype)
                arrays_grp = ptype_grp.create_group('arrays')
                data = self.all_array_data[ptype]
                self._set_constants(pdata, ptype_grp)
                self._set_properties(pdata, arrays_grp, data)
            self._set_solver_data(solver_grp)

    def _load(self, fname):
        if has_h5py():
            import h5py
        else:
            msg = "Install python-h5py to load this file"
            raise ImportError(msg)

        ret = {}
        with h5py.File(fname, 'r') as f:
            solver_grp = f['solver_data']
            particles_grp = f['particles']
            ret["solver_data"] = self._get_solver_data(solver_grp)
            ret["arrays"] = self._get_particles(particles_grp)
        return ret

    def _get_particles(self, grp):

        particles = {}
        for name, prop_array in grp.items():
            output_array = []
            const_grp = prop_array['constants']
            arrays_grp = prop_array['arrays']
            constants = self._get_constants(const_grp)
            array = ParticleArray(_to_str(name), constants=constants)

            for pname, h5obj in arrays_grp.items():
                prop_name = _to_str(h5obj.attrs['name'])
                type_ = _to_str(h5obj.attrs['type'])
                default = h5obj.attrs['default']
                stride = h5obj.attrs.get('stride', 1)
                if h5obj.attrs['stored']:
                    output_array.append(_to_str(pname))
                    array.add_property(
                        prop_name, type=type_, default=default,
                        data=numpy.array(h5obj),
                        stride=stride
                    )
                else:
                    array.add_property(prop_name, type=type_, stride=stride)
            array.set_output_arrays(output_array)
            particles[str(name)] = array
        return particles

    def _get_solver_data(self, grp):
        solver_data = {}
        for name, value in grp.attrs.items():
            solver_data[_to_str(name)] = value
        return solver_data

    def _get_constants(self, grp):
        constants = {}
        for const_name, const_data in grp.items():
            constants[_to_str(const_name)] = numpy.array(const_data)
        return constants

    def _get_compress_options(self):
        if self.compress:
            return dict(compression="gzip",
                        compression_opts=COMPRESSION_LEVEL)
        else:
            return {}

    def _set_constants(self, pdata, ptype_grp):
        pconstants = pdata['constants']
        constGroup = ptype_grp.create_group('constants')
        c_kw = self._get_compress_options()
        for constName, constArray in pconstants.items():
            constGroup.create_dataset(constName, data=constArray, **c_kw)

    def _set_properties(self, pdata, ptype_grp, data):
        c_kw = self._get_compress_options()
        for propname, attributes in pdata['properties'].items():
            if propname in data:
                array = data[propname]
                prop = ptype_grp.create_dataset(propname, data=array, **c_kw)
                prop.attrs['stored'] = True
            else:
                prop = ptype_grp.create_dataset(propname, (0,))
                prop.attrs['stored'] = False

            for attname, value in attributes.items():
                if value is None:
                    value = 'None'
                prop.attrs[attname] = value

    def _set_solver_data(self, grp):
        for name, data in self.solver_data.items():
            grp.attrs[name] = data


def load(fname):
    """
    Load the output data

    Parameters
    ----------
    fname: str
        Name of the file or full path


    Examples
    --------
    >>> data = load('elliptical_drop_100.npz')
    >>> data.keys()
    ['arrays', 'solver_data']
    >>> arrays = data['arrays']
    >>> arrays.keys()
    ['fluid']
    >>> fluid = arrays['fluid']
    >>> type(fluid)
    pysph.base.particle_array.ParticleArray
    >>> data['solver_data']
    {'count': 100, 'dt': 4.6416394784204199e-05, 't': 0.0039955855395528766}
    """

    if fname.endswith('npz'):
        output = NumpyOutput()
    elif fname.endswith('hdf5'):
        output = HDFOutput()
    if os.path.isfile(fname):
        return output.load(fname)
    else:
        msg = "File not present"
        raise RuntimeError(msg)


def dump(filename, particles, solver_data, detailed_output=False,
         only_real=True, mpi_comm=None, compress=False):

    """
    Dump the given particles and solver data to the given filename.

    Parameters
    ----------

    filename: str
        Filename to dump to.

    particles: sequence(ParticleArray)
        Sequence of particle arrays to dump.

    solver_data: dict
        Additional information to dump about solver state.

    detailed_output: bool
        Specifies if all arrays should be dumped.

    only_real: bool
        Only dump the real particles.

    mpi_comm: mpi4pi.MPI.Intracomm
        An MPI communicator to use for parallel commmunications.

    compress: bool
        Specify if the  file is to be compressed or not.

    If `mpi_comm` is not passed or is set to None the local particles alone
    are dumped, otherwise only rank 0 dumps the output.

    """
    if filename.endswith(output_formats):
        fname = os.path.splitext(filename)[0]
    else:
        fname = filename
        filename = fname + '.hdf5'
    if filename.endswith('hdf5') and has_h5py():
        file_format = 'hdf5'
        output = HDFOutput(detailed_output, only_real, mpi_comm, compress)
    else:
        output = NumpyOutput(detailed_output, only_real, mpi_comm, compress)
        file_format = 'npz'
    filename = fname + '.' + file_format
    output.dump(filename, particles, solver_data)



'''
Functions can be used to generate points to describe a given input
mesh.

Supported mesh formats: All file formats supported by meshio.
(https://github.com/nschloe/meshio)
'''

import numpy as np
import meshio
from pysph.tools.mesh_tools import surface_points, surf_points_uniform


class Mesh:
    def __init__(self, file_name, file_type=None):
        if file_type is None:
            self.mesh = meshio.read(file_name)
        else:
            self.mesh = meshio.read(file_name, file_type)

        self.cells = np.array([], dtype=int).reshape(0, 3)

    def extract_connectivity_info(self):
        cell_blocks = self.mesh.cells
        for block in cell_blocks:
            self.cells = np.concatenate((self.cells, block.data))

        return self.cells

    def extract_coordinates(self):
        x, y, z = self.mesh.points.T
        self.x, self.y, self.z = x, y, z

        return x, y, z

    def compute_normals(self):
        n = self.cells.shape[0]
        self.normals = np.zeros((n, 3))
        points = self.mesh.points

        idx = self.cells
        pts = points[idx]
        a = pts[:, 1] - pts[:, 0]
        b = pts[:, 2] - pts[:, 0]

        normals = np.cross(a, b)
        mag = np.linalg.norm(normals, axis=1)
        mag.shape = (n, 1)
        self.normals = normals/mag

        return self.normals


def mesh2points(file_name, dx, file_format=None, uniform=False):
    '''
    Generates points with a spacing dx to describe the surface of the
    input mesh file.

    Supported file formats: Refer to https://github.com/nschloe/meshio

    Only works with triangle meshes.

    Parameters
    ----------

    file_name : string
        Mesh file name
    dx : float
        Required spacing between generated particles
    file_format : str
        Mesh file format
    uniform : bool
        If True generates points on a grid of spacing dx

    Returns
    -------
    xf, yf, zf : ndarray
        1d numpy arrays with x, y, z coordinates of covered surface
    '''
    mesh = Mesh(file_name)
    cells = mesh.extract_connectivity_info()
    x, y, z = mesh.extract_coordinates()

    if uniform is False:
        xf, yf, zf = surface_points(x, y, z, cells, dx)

    else:
        if file_format is 'stl':
            normals = mesh.mesh.cell_data['facet_normals'][0]
        else:
            normals = mesh.compute_normals()

        xf, yf, zf = surf_points_uniform(x, y, z, cells, normals, dx, dx)

    return xf, yf, zf



# cython: language_level=3, embedsignature=True
# distutils: language=c++
from libcpp.vector cimport vector

from .nnps_base cimport *

ctypedef unsigned int u_int

cdef extern from 'math.h':
    int abs(int) nogil
    double ceil(double) nogil
    double floor(double) nogil
    double fabs(double) nogil
    double fmax(double, double) nogil
    double fmin(double, double) nogil

#Imports for SpatialHashNNPS
cdef extern from "spatial_hash.h":
    cdef cppclass HashEntry:
        double h_max

        vector[unsigned int] *get_indices() nogil

    cdef cppclass HashTable:
        long long int table_size

        HashTable(long long int) nogil except +
        void add(int, int, int, int, double) nogil
        HashEntry* get(int, int, int) nogil
        int number_of_particles() nogil

cdef class StratifiedHashNNPS(NNPS):
    ############################################################################
    # Data Attributes
    ############################################################################
    cdef long long int table_size               # Size of hashtable
    cdef double radius_scale2

    cdef public int num_levels
    cdef public int H

    cdef double interval_size

    cdef HashTable*** hashtable
    cdef HashTable** current_hash

    cdef double** cell_sizes
    cdef double* current_cells

    cdef NNPSParticleArrayWrapper dst, src

    ##########################################################################
    # Member functions
    ##########################################################################

    cpdef set_context(self, int src_index, int dst_index)

    cpdef int count_particles(self, int interval)

    cpdef double get_binning_size(self, int interval)

    cdef void find_nearest_neighbors(self, size_t d_idx, UIntArray nbrs) nogil

    cdef inline int _h_mask_exact(self, int* x, int* y, int* z, int H) nogil

    cdef inline int _neighbor_boxes(self, int i, int j, int k,
            int* x, int* y, int* z, int H) nogil

    cdef inline int _get_hash_id(self, double h) nogil

    cdef inline void _set_h_max(self, double* current_cells, double* src_h_ptr,
            int num_particles) nogil

    cdef inline double _get_h_max(self, double* current_cells, int hash_id) nogil

    cpdef _refresh(self)

    cpdef _bin(self, int pa_index, UIntArray indices)



#cython: embedsignature=True

"""A handy set of classes/functions for 3D points.
"""

from cpython cimport *
# numpy imports
cimport numpy
import numpy

# IntPoint's maximum absolute value must be less than `IntPoint_maxint`
# this is due to the hash implementation
cdef int IntPoint_maxint = 2**20

DTYPE = numpy.float64
ctypedef numpy.float_t DTYPE_t


###############################################################################
# `Point` class.
###############################################################################
cdef class Point:
    """
    This class represents a Point in 3D space.
    """

    # Declared in the .pxd file.
    #cdef public double x, y, z

    property x:
        def __get__(Point self):
            return self.data.x
        def __set__(Point self, double x):
            self.data.x = x

    property y:
        def __get__(Point self):
            return self.data.y
        def __set__(Point self, double y):
            self.data.y = y

    property z:
        def __get__(Point self):
            return self.data.z
        def __set__(Point self, double z):
            self.data.z = z

    ######################################################################
    # `object` interface.
    ######################################################################
    def __init__(Point self, double x=0.0, double y=0.0, double z=0.0):
        """Constructor for a Point."""
        self.data.x = x
        self.data.y = y
        self.data.z = z

    def __reduce__(Point self):
        """
        Implemented to facilitate pickling of the Point extension type.
        """
        d = {}
        d['x'] = self.data.x
        d['y'] = self.data.y
        d['z'] = self.data.z

        return (Point, (), d)

    def __setstate__(Point self, d):
        self.data.x = d['x']
        self.data.y = d['y']
        self.data.z = d['z']

    def __str__(Point self):
        return '(%f, %f, %f)'%(self.data.x, self.data.y, self.data.z)

    def __repr__(Point self):
        return 'Point(%g, %g, %g)'%(self.data.x, self.data.y, self.data.z)

    def __add__(Point self, Point p):
        return Point_new(self.data.x + p.data.x, self.data.y + p.data.y,
                         self.data.z + p.data.z)

    def __sub__(Point self, Point p):
        return Point_new(self.data.x - p.data.x, self.data.y - p.data.y,
                         self.data.z - p.data.z)

    def __mul__(Point self, double m):
        return Point_new(self.data.x*m, self.data.y*m, self.data.z*m)

    def __div__(Point self, double m):
        return Point_new(self.data.x/m, self.data.y/m, self.data.z/m)

    def __abs__(Point self):
        return cPoint_length(self.data)

    def __neg__(Point self):
        return Point_new(-self.data.x, -self.data.y, -self.data.z)

    def __richcmp__(Point self, Point p, int oper):
        if oper == 2: # ==
            if self.data.x == p.data.x and self.data.y == p.data.y and self.data.z == p.data.z:
                return True
            return False
        elif oper == 3: # !=
            if self.data.x == p.data.x and self.data.y == p.data.y and self.data.z == p.data.z:
                return False
            return True
        else:
            raise TypeError('No ordering is possible for Points.')

    def __iadd__(Point self, Point p):
        self.data.x += p.data.x
        self.data.y += p.data.y
        self.data.z += p.data.z
        return self

    def __isub__(Point self, Point p):
        self.data.x -= p.data.x
        self.data.y -= p.data.y
        self.data.z -= p.data.z
        return self

    def __imul__(Point self, double m):
        self.data.x *= m
        self.data.y *= m
        self.data.z *= m
        return self

    def __idiv__(Point self, double m):
        self.data.x /= m
        self.data.y /= m
        self.data.z /= m
        return self

    ######################################################################
    # `Point` interface.
    ######################################################################
    cpdef set(Point self, double x, double y, double z):
        """Set the position from a given array.
        """
        self.data.x = x
        self.data.y = y
        self.data.z = z

    cdef set_from_cPoint(Point self, cPoint value):
        self.data.x = value.x
        self.data.y = value.y
        self.data.z = value.z

    cpdef numpy.ndarray asarray(Point self):
        """Return a numpy array with the coordinates."""
        cdef numpy.ndarray[DTYPE_t, ndim=1] r = numpy.empty(3)
        r[0] = self.data.x
        r[1] = self.data.y
        r[2] = self.data.z
        return r

    cpdef double norm(Point self):
        """Return the square of the Euclidean distance to this point."""
        return cPoint_norm(self.data)

    cpdef double length(Point self):
        """Return the Euclidean distance to this point."""
        return cPoint_length(self.data)

    cpdef double dot(Point self, Point p):
        """Return the dot product of this point with another."""
        return cPoint_dot(self.data, p.data)

    cpdef Point cross(Point self, Point p):
        """Return the cross product of this point with another, i.e.
        `self` cross `p`."""
        return Point_new(self.data.y*p.data.z - self.data.z*p.data.y,
                         self.data.z*p.data.x - self.data.x*p.data.z,
                         self.data.x*p.data.y - self.data.y*p.data.x)

    cpdef double distance(Point self, Point p):
        """Return the distance between this point and p"""
        return cPoint_distance(self.data, p.data)

    cdef cPoint to_cPoint(Point self):
        return self.data

    def normalize(self):
        """ Normalize the point """
        cdef double norm = cPoint_length(self.data)

        self.data.x /= norm
        self.data.y /= norm
        self.data.z /= norm

cdef class IntPoint:

    property x:
        def __get__(self):
            return self.data.x

    property y:
        def __get__(self):
            return self.data.y

    property z:
        def __get__(self):
            return self.data.z

    def __init__(self, int x=0, int y=0, int z=0):
        self.data.x = x
        self.data.y = y
        self.data.z = z

    def __reduce__(self):
        """
        Implemented to facilitate pickling of the IntPoint extension type.
        """
        d = {}
        d['x'] = self.data.x
        d['y'] = self.data.y
        d['z'] = self.data.z

        return (IntPoint, (), d)

    def __setstate__(self, d):
        self.data.x = d['x']
        self.data.y = d['y']
        self.data.z = d['z']

    def __str__(self):
        return '(%d,%d,%d)'%(self.data.x, self.data.y, self.data.z)

    def __repr__(self):
        return 'IntPoint(%d,%d,%d)'%(self.data.x, self.data.y, self.data.z)

    cdef IntPoint copy(self):
        return IntPoint_new(self.data.x, self.data.y, self.data.z)

    cpdef numpy.ndarray asarray(self):
        cdef numpy.ndarray[ndim=1,dtype=numpy.int_t] arr = numpy.empty(3,
                                                            dtype=numpy.int)
        arr[0] = self.data.x
        arr[1] = self.data.y
        arr[2] = self.data.z

        return arr

    cdef bint is_equal(self, IntPoint p):
        return cIntPoint_is_equal(self.data, p.data)

    cdef IntPoint diff(self, IntPoint p):
        return IntPoint_new(self.data.x-p.data.x, self.data.y-p.data.y,
                            self.data.z-p.data.z)

    cdef tuple to_tuple(self):
        cdef tuple t = (self.data.x, self.data.y, self.data.z)
        return t

    def __richcmp__(IntPoint self, IntPoint p, int oper):
        if oper == 2: # ==
            return cIntPoint_is_equal(self.data, p.data)
        elif oper == 3: # !=
            return not cIntPoint_is_equal(self.data, p.data)
        else:
            raise TypeError('No ordering is possible for Points.')

    def __hash__(self):
        cdef long ret = self.data.x + IntPoint_maxint
        ret = 2 * IntPoint_maxint * ret + self.data.y + IntPoint_maxint
        return 2 * IntPoint_maxint * ret + self.data.z + IntPoint_maxint

    def py_is_equal(self, IntPoint p):
        return self.is_equal(p)

    def py_diff(self, IntPoint p):
        return self.diff(p)

    def py_copy(self):
        return self.copy()




# Standard library imports.
import numpy
from textwrap import dedent
import unittest

# Local imports.
from compyle.api import KnownType
from pysph.sph.equation import (
    BasicCodeBlock, Context, CythonGroup, Equation, Group, sort_precomputed
)


class TestContext(unittest.TestCase):
    def test_basic_usage(self):
        c = Context(a=1, b=2)
        self.assertEqual(c.a, 1)
        self.assertEqual(c.b, 2)
        self.assertEqual(c['a'], 1)
        self.assertEqual(c['b'], 2)
        c.c = 3
        self.assertEqual(c.c, 3)
        self.assertEqual(c['c'], 3)

    def test_context_behaves_like_dict(self):
        c = Context(a=1)
        c.b = 2
        keys = list(c.keys())
        keys.sort()
        self.assertEqual(keys, ['a', 'b'])
        values = list(c.values())
        values.sort()
        self.assertEqual(values, [1, 2])
        self.assertTrue('a' in c)
        self.assertTrue('b' in c)
        self.assertTrue('c' not in c)


class TestBase(unittest.TestCase):
    def assert_seq_equal(self, got, expect):
        g = list(got)
        g.sort()
        e = list(expect)
        e.sort()
        self.assertEqual(g, e, 'got %s, expected %s' % (g, e))


class TestBasicCodeBlock(TestBase):

    def test_basic_code_block(self):
        code = '''
        x = 1
        d_x[d_idx] += s_x[s_idx] + x
        '''
        cb = BasicCodeBlock(code=code)
        expect = ['d_idx', 'd_x', 's_idx', 's_x', 'x']
        self.assert_seq_equal(cb.symbols, expect)
        self.assert_seq_equal(cb.src_arrays, ['s_x'])
        self.assert_seq_equal(cb.dest_arrays, ['d_x'])
        ctx = cb.context
        self.assertEqual(ctx.s_idx, 0)
        self.assertEqual(ctx.d_idx, 0)
        x = numpy.zeros(2, dtype=float)
        self.assertTrue(numpy.alltrue(ctx.d_x == x))
        self.assertTrue(numpy.alltrue(ctx.s_x == x))

    def test_that_code_block_is_callable(self):
        code = '''
        x = 1
        d_x[d_idx] += s_x[s_idx] + x
        '''
        cb = BasicCodeBlock(code=code)
        # The code block should be callable.
        res = cb()
        self.assertTrue(sum(res.d_x) == 1)

        # Should take arguments to update the context.
        x = numpy.ones(2, dtype=float)*10
        res = cb(s_x=x)
        self.assertTrue(sum(res.d_x) == 11)


class TestEquations(TestBase):
    def test_simple_equation(self):
        eq = Equation('fluid', None)
        self.assertEqual(eq.name, 'Equation')
        self.assertEqual(eq.no_source, True)
        self.assertEqual(eq.dest, 'fluid')
        self.assertEqual(eq.sources, None)
        self.assertFalse(hasattr(eq, 'loop'))
        self.assertFalse(hasattr(eq, 'post_loop'))
        self.assertFalse(hasattr(eq, 'initialize'))

        eq = Equation('fluid', sources=['fluid'])
        self.assertEqual(eq.name, 'Equation')
        self.assertEqual(eq.no_source, False)
        self.assertEqual(eq.dest, 'fluid')
        self.assertEqual(eq.sources, ['fluid'])

        class Test(Equation):
            pass
        eq = Test('fluid', [])
        self.assertEqual(eq.name, 'Test')
        self.assertEqual(eq.no_source, True)

    def test_continuity_equation(self):
        from pysph.sph.basic_equations import ContinuityEquation
        e = ContinuityEquation(dest='fluid', sources=['fluid'])
        # Call the loop code.

        d_arho = [0.0, 0.0, 0.0]
        s_m = [0.0, 0.0]
        e.loop(d_idx=0, d_arho=d_arho, s_idx=0, s_m=s_m,
               DWIJ=[0, 0, 0], VIJ=[0, 0, 0])
        self.assertEqual(d_arho[0], 0.0)
        self.assertEqual(d_arho[1], 0.0)
        # Now call with specific arguments.
        s_m = [1, 1]
        e.loop(d_idx=0, d_arho=d_arho, s_idx=0, s_m=s_m,
               DWIJ=[1, 1, 1], VIJ=[1, 1, 1])
        self.assertEqual(d_arho[0], 3.0)
        self.assertEqual(d_arho[1], 0.0)

    def test_order_of_precomputed(self):
        try:
            pre_comp = Group.pre_comp
            pre_comp.AIJ = BasicCodeBlock(
                code=dedent("""
                AIJ[0] = XIJ[0]/RIJ
                AIJ[1] = XIJ[1]/RIJ
                AIJ[2] = XIJ[2]/RIJ
                """),
                AIJ=[1.0, 0.0, 0.0])
            input = dict((x, pre_comp[x]) for x in ['RIJ', 'R2IJ',
                                                    'XIJ', 'HIJ', 'AIJ'])
            pre = sort_precomputed(input, pre_comp)
            self.assertEqual(
                list(pre.keys()), ['HIJ', 'XIJ', 'R2IJ', 'RIJ', 'AIJ']
            )
        finally:
            from pysph.sph.equation import precomputed_symbols
            Group.pre_comp = precomputed_symbols()


class Equation1(Equation):
    def loop(self, WIJ=0.0):
        x = WIJ
        x += 1

    def post_loop(self, d_idx, d_h):
        x = d_h[d_idx]
        x += 1


class Equation2(Equation):
    def loop(self, d_idx, s_idx):
        x = s_idx + d_idx
        x += 1


class TestGroup(TestBase):
    def setUp(self):
        from pysph.sph.basic_equations import SummationDensity
        from pysph.sph.wc.basic import TaitEOS
        self.group = CythonGroup(
            [SummationDensity('f', ['f']),
             TaitEOS('f', None, rho0=1.0, c0=1.0, gamma=1.4, p0=1.0)]
        )

    def test_precomputed(self):
        g = self.group
        self.assertEqual(len(g.precomputed), 5)
        self.assertEqual(list(g.precomputed.keys()),
                         ['HIJ', 'XIJ', 'R2IJ', 'RIJ', 'WIJ'])

    def test_array_names(self):
        g = self.group
        src, dest = g.get_array_names()
        s_ex = ['s_m', 's_x', 's_y', 's_z', 's_h']
        d_ex = ['d_rho', 'd_p', 'd_h', 'd_cs', 'd_x', 'd_y', 'd_z']
        self.assert_seq_equal(src, s_ex)
        self.assert_seq_equal(dest, d_ex)

    def test_variable_names(self):
        g = self.group
        names = g.get_variable_names()
        expect = ['WIJ', 'RIJ', 'R2IJ', 'XIJ', 'HIJ']
        self.assert_seq_equal(names, expect)

    def test_array_declarations(self):
        g = self.group
        expect = 'cdef double* d_x'
        self.assertEqual(g.get_array_declarations(['d_x']), expect)

    def test_array_declarations_with_known_types(self):
        # Given
        g = self.group
        known_types = {'d_x': KnownType('float*')}
        # When
        result = g.get_array_declarations(['d_x'], known_types)
        # Then.
        expect = 'cdef float* d_x'
        self.assertEqual(result, expect)

    def test_variable_declarations(self):
        g = self.group
        context = Context(x=1.0)
        expect = 'cdef double x = 1.0'
        self.assertEqual(g.get_variable_declarations(context), expect)
        context = Context(x=1)
        expect = 'cdef long x = 1'
        self.assertEqual(g.get_variable_declarations(context), expect)

        context = Context(x=[1., 2.])
        expect = ('cdef DoubleArray _x = DoubleArray(aligned(2, 8)*'
                  'self.n_threads)\n'
                  'cdef double* x = _x.data')
        self.assertEqual(g.get_variable_declarations(context), expect)

        context = Context(x=(0, 1., 2.))
        expect = ('cdef DoubleArray _x = DoubleArray(aligned(3, 8)*'
                  'self.n_threads)\n'
                  'cdef double* x = _x.data')
        self.assertEqual(g.get_variable_declarations(context), expect)

    def test_loop_code(self):
        from pysph.base.kernels import CubicSpline
        k = CubicSpline(dim=3)
        e1 = Equation1('f', ['f'])
        e2 = Equation2('f', ['f'])
        g = CythonGroup([e1, e2])
        # First get the equation wrappers so the equation names are setup.
        g.get_equation_wrappers()
        result = g.get_loop_code(k)
        expect = dedent('''\
            HIJ = 0.5*(d_h[d_idx] + s_h[s_idx])
            XIJ[0] = d_x[d_idx] - s_x[s_idx]
            XIJ[1] = d_y[d_idx] - s_y[s_idx]
            XIJ[2] = d_z[d_idx] - s_z[s_idx]
            R2IJ = XIJ[0]*XIJ[0] + XIJ[1]*XIJ[1] + XIJ[2]*XIJ[2]
            RIJ = sqrt(R2IJ)
            WIJ = self.kernel.kernel(XIJ, RIJ, HIJ)

            self.equation10.loop(WIJ)
            self.equation20.loop(d_idx, s_idx)
            ''')
        msg = 'EXPECTED:\n%s\nGOT:\n%s' % (expect, result)
        self.assertEqual(result, expect, msg)

    def test_post_loop_code(self):
        from pysph.base.kernels import CubicSpline
        k = CubicSpline(dim=3)
        e1 = Equation1('f', ['f'])
        e2 = Equation2('f', ['f'])
        g = CythonGroup([e1, e2])
        # First get the equation wrappers so the equation names are setup.
        g.get_equation_wrappers()
        result = g.get_post_loop_code(k)
        expect = dedent('''\
            self.equation10.post_loop(d_idx, d_h)
            ''')
        msg = 'EXPECTED:\n%s\nGOT:\n%s' % (expect, result)
        self.assertEqual(result, expect, msg)


if __name__ == '__main__':
    unittest.main()



"""Incompressible flow past a periodic array of cylinders. (42 hours)


See Ellero and Adams, International Journal for Numerical Methods in
Engineering, 2011, vol 86, pp 1027-1040 for the detailed parameters for this
problem and also  Adami, Hu and Adams, JCP, 2013, vol 241, pp 292-307.

In particular, we note that we set c0 from Ellero and Adams as using the
value from Adami et al. will cause the solution to blow up.

If one sets c0=10*Umax and sets pb=300*p0, that will cause particles to void
at the rear of the cylinder.

"""

import os
import numpy as np

# PySPH imports
from pysph.base.nnps import DomainManager
from pysph.base.utils import get_particle_array
from pysph.solver.application import Application

from pysph.sph.scheme import TVFScheme


# domain and reference values
L = 0.12
Umax = 1.2e-4
a = 0.02
H = 4*a
fx = 2.5e-4

# c0 is set from Ellero and Adams.
# Note that setting this to 0.1*np.sqrt(a*fx) as per Adami Hu and Adams is
# incorrect and will actually cause a blow up of the solution.
c0 = 0.02
rho0 = 1000.0
p0 = c0*c0*rho0
pb = p0

# Reynolds number and kinematic viscosity
nu = 0.1/rho0
Re = a*Umax/nu

# Numerical setup
nx = 144
dx = L/nx
ghost_extent = 5 * 1.5 * dx
hdx = 1.2

# adaptive time steps
h0 = hdx * dx
dt_cfl = 0.25 * h0/(c0 + Umax)
dt_viscous = 0.125 * h0**2/nu
dt_force = 0.25 * np.sqrt(h0/abs(fx))

T = a/Umax

tf = 2.5*T
dt = min(dt_cfl, dt_viscous, dt_force)


class PeriodicCylinders(Application):

    def create_domain(self):
        # domain for periodicity
        domain = DomainManager(xmin=0, xmax=L, periodic_in_x=True)
        return domain

    def create_particles(self):
        # create all the particles
        _x = np.arange(dx/2, L, dx)
        _y = np.arange(-ghost_extent, H+ghost_extent, dx)
        x, y = np.meshgrid(_x, _y)
        x = x.ravel()
        y = y.ravel()

        # sort out the fluid and the solid
        indices = []
        cx = 0.5 * L
        cy = 0.5 * H
        for i in range(x.size):
            xi = x[i]
            yi = y[i]
            if (np.sqrt((xi-cx)**2 + (yi-cy)**2) > a):
                if ((yi > 0) and (yi < H)):
                    indices.append(i)

        # create the arrays
        solid = get_particle_array(name='solid', x=x, y=y)

        # remove the fluid particles from the solid
        fluid = solid.extract_particles(indices)
        fluid.set_name('fluid')
        solid.remove_particles(indices)

        print("Periodic cylinders :: Re = %g, nfluid = %d, nsolid=%d, dt = %g"
              % (Re, fluid.get_number_of_particles(),
                 solid.get_number_of_particles(), dt))
        print("tf = %f" % tf)

        # add requisite properties to the arrays:
        self.scheme.setup_properties([fluid, solid])

        # setup the particle properties
        volume = dx * dx

        # mass is set to get the reference density of rho0
        fluid.m[:] = volume * rho0
        solid.m[:] = volume * rho0

        # initial particle density
        fluid.rho[:] = rho0
        solid.rho[:] = rho0

        # volume is set as dx^2. V is the number density form of the
        # particle volume and will be computed in the equations for the
        # fluid phase. The initial values are used for the solid phase
        fluid.V[:] = 1./volume
        solid.V[:] = 1./volume

        # particle smoothing lengths
        fluid.h[:] = hdx * dx
        solid.h[:] = hdx * dx

        # return the particle list
        return [fluid, solid]

    def create_scheme(self):
        s = TVFScheme(
            ['fluid'], ['solid'], dim=2, rho0=rho0, c0=c0, nu=nu,
            p0=p0, pb=p0, h0=dx*hdx, gx=fx
        )
        s.configure_solver(tf=tf, dt=dt, n_damp=100, pfreq=500)
        return s

    def post_process(self, info_fname):
        self.read_info(info_fname)
        if len(self.output_files) == 0:
            return

        t, cd = self._plot_cd_vs_t()
        res = os.path.join(self.output_dir, 'results.npz')
        np.savez(res, t=t, cd=cd)

    def _plot_cd_vs_t(self):
        from pysph.solver.utils import iter_output, load
        from pysph.tools.sph_evaluator import SPHEvaluator
        from pysph.sph.equation import Group
        from pysph.base.kernels import QuinticSpline
        from pysph.sph.wc.transport_velocity import (
            SetWallVelocity, MomentumEquationPressureGradient,
            SolidWallNoSlipBC, SolidWallPressureBC, VolumeSummation
        )

        data = load(self.output_files[0])
        solid = data['arrays']['solid']
        fluid = data['arrays']['fluid']
        x, y = solid.x.copy(), solid.y.copy()
        cx = 0.5 * L
        cy = 0.5 * H
        inside = np.sqrt((x-cx)**2 + (y-cy)**2) <= a
        dest = solid.extract_particles(inside.nonzero()[0])
        # We use the same equations for this as the simulation, except that we
        # do not include the acceleration terms as these are externally
        # imposed.  The goal of these is to find the force of the fluid on the
        # cylinder, thus, gx=0.0 is used in the following.
        equations = [
            Group(
                equations=[
                    VolumeSummation(
                        dest='fluid', sources=['fluid', 'solid']
                    ),
                    VolumeSummation(
                        dest='solid', sources=['fluid', 'solid']
                    ),
                    ], real=False),

            Group(
                equations=[
                    SetWallVelocity(dest='solid', sources=['fluid']),
                    ], real=False),

            Group(
                equations=[
                    SolidWallPressureBC(dest='solid', sources=['fluid'],
                                        gx=0.0, b=1.0, rho0=rho0, p0=p0),
                    ], real=False),

            Group(
                equations=[
                    # Pressure gradient terms
                    MomentumEquationPressureGradient(
                        dest='fluid', sources=['solid'], gx=0.0, pb=pb),
                    SolidWallNoSlipBC(
                        dest='fluid', sources=['solid'], nu=nu),
                    ], real=True),
        ]

        sph_eval = SPHEvaluator(
            arrays=[dest, fluid], equations=equations, dim=2,
            kernel=QuinticSpline(dim=2)
        )

        t, cd = [], []
        for sd, fluid in iter_output(self.output_files, 'fluid'):
            fluid.remove_property('vmag2')
            t.append(sd['t'])
            sph_eval.update_particle_arrays([dest, fluid])
            sph_eval.evaluate()
            Fx = np.sum(-fluid.au*fluid.m)
            cd.append(Fx/(nu*rho0*Umax))

        t, cd = list(map(np.asarray, (t, cd)))

        # Now plot the results.
        import matplotlib
        matplotlib.use('Agg')

        from matplotlib import pyplot as plt
        plt.figure()
        plt.plot(t, cd)
        plt.xlabel('$t$')
        plt.ylabel(r'$C_D$')
        fig = os.path.join(self.output_dir, "cd_vs_t.png")
        plt.savefig(fig, dpi=300)
        plt.close()

        return t, cd


if __name__ == '__main__':
    app = PeriodicCylinders()
    app.run()
    app.post_process(app.info_filename)



"""
SPHysics case2 - dambreak on wet surface (5 minutes)
"""

from pysph.base.kernels import CubicSpline
from pysph.solver.application import Application

from pysph.sph.integrator_step import WCSPHStep
from pysph.base.utils import get_particle_array

import numpy as np
from pysph.sph.scheme import AdamiHuAdamsScheme, WCSPHScheme, SchemeChooser
from pysph.sph.wc.edac import EDACScheme
from pysph.tools.geometry import remove_overlap_particles, rotate
from pysph.tools.geometry import get_2d_tank, get_2d_block


def get_dam_geometry(dx_tank=0.03, dx_fluid=0.03, r_tank=100.0, h_f=2.0,
                     l_f=1.0, r_fluid=100.0, hdx=1.5, l_tank=4.0,
                     h_tank=4.0, h_f2=1.0):
    tank_x, tank_y = get_2d_tank(dx_tank, length=l_tank, height=h_tank,
                                 num_layers=4)
    rho_tank = np.ones_like(tank_x) * r_tank
    m_tank = rho_tank * dx_tank * dx_tank
    h_t = np.ones_like(tank_x) * dx_tank * hdx
    tank = get_particle_array(name='dam', x=tank_x, y=tank_y, h=h_t,
                              rho=rho_tank, m=m_tank)
    center = np.array([(l_f - l_tank) / 2.0, h_f / 2.0])
    fluid_x1, fluid_y1 = get_2d_block(dx_fluid, l_f, h_f, center)
    center = np.array([l_f / 2.0, h_f2 / 2.0])
    fluid_x2, fluid_y2 = get_2d_block(dx_fluid, l_tank - l_f - 2.0 * dx_fluid,
                                      h_f2, center)
    fluid_x = np.concatenate([fluid_x1, fluid_x2])
    fluid_y = np.concatenate([fluid_y1, fluid_y2])
    h_fluid = np.ones_like(fluid_x) * dx_fluid * hdx
    r_f = np.ones_like(fluid_x) * r_fluid
    m_f = r_f * dx_fluid * dx_fluid
    fluid = get_particle_array(name='fluid', x=fluid_x, y=fluid_y,
                               h=h_fluid, rho=r_f, m=m_f)
    remove_overlap_particles(fluid, tank, dx_tank, 2)
    return fluid, tank


class Dambreak_2D(Application):

    def add_user_options(self, group):
        group.add_argument(
            "--hdx", action="store", type=float, dest="hdx", default=1.3,
            help="h/dx value used in SPH to change the smoothness")
        group.add_argument(
            "--dx", action="store", type=float, dest="dx", default=0.005,
            help="spacing between the particles")

    def consume_user_options(self):
        self.hdx = self.options.hdx
        self.dx = self.options.dx
        self.h0 = self.hdx * self.dx
        self.dt = 0.15 * self.h0 / co

    def create_particles(self):
        fluid, dam = get_dam_geometry(self.dx, self.dx, hdx=self.hdx,
                                      h_f=h_fluid, h_f2=h_fluid2, r_fluid=ro,
                                      r_tank=ro, l_f=l_fluid, l_tank=l_dam,
                                      h_tank=h_dam)
        self.scheme.setup_properties([fluid, dam])
        particles = [fluid, dam]
        return particles

    def create_scheme(self):
        aha = AdamiHuAdamsScheme(['fluid'], ['dam'], dim=2, rho0=ro, c0=co,
                                 alpha=alp, gy=-9.81, nu=0.0, h0=0.005,
                                 gamma=1.0)
        wcsph = WCSPHScheme(['fluid'], ['dam'], dim=2, rho0=ro, c0=co,
                            h0=0.005, hdx=1.3, hg_correction=True, gy=-9.81,
                            alpha=alp, gamma=gamma, update_h=True)
        edac = EDACScheme(['fluid'], ['dam'], dim=2, rho0=ro, c0=co, gy=-9.81,
                          alpha=0.0, nu=0.0, h=0.005, clamp_p=True)
        return SchemeChooser(default='wcsph', wcsph=wcsph, aha=aha, edac=edac)

    def configure_scheme(self):
        s = self.scheme
        scheme = self.options.scheme
        if scheme == 'wcsph':
            s.configure(h0=self.h0, hdx=self.hdx)
        elif scheme == 'aha':
            s.configure(h0=self.h0)
        elif scheme == 'edac':
            s.configure(h=self.h0)
        s.configure_solver(kernel=CubicSpline(dim=2), dt=self.dt, tf=1.2,
                           adaptive_timestep=False)


if __name__ == '__main__':
    l_dam = 2.0
    h_dam = 0.16
    h_fluid = 0.15
    l_fluid = 0.376
    h_fluid2 = 0.018
    gamma = 7.0
    alp = 0.2
    ro = 100.0
    co = 10.0 * np.sqrt(2.0 * 9.81 * h_fluid)
    app = Dambreak_2D()
    app.run()



#cython: embedsignature=True

# Numpy
import numpy as np
cimport numpy as np

# MPI4PY
import mpi4py.MPI as mpi

from cpython.list cimport PyList_Append, PyList_GET_SIZE

from compyle.profile import profile_ctx, ProfileContext

# PyZoltan
from pyzoltan.czoltan cimport czoltan
from pyzoltan.czoltan.czoltan cimport Zoltan_Struct
from pyzoltan.core import zoltan_utils

# PySPH imports
from pysph.base.nnps_base cimport DomainManager, Cell, find_cell_id, arange_uint
from pysph.base.utils import ParticleTAGS

cdef int Local = ParticleTAGS.Local
cdef int Remote = ParticleTAGS.Remote

cdef extern from 'math.h':
    cdef double ceil(double) nogil
    cdef double floor(double) nogil
    cdef double fabs(double) nogil


cpdef get_strided_indices(np.ndarray indices, int stride):
    cdef int j
    cdef list args = []
    if stride == 1:
        return indices
    else:
        tmp = indices*stride
        for j in range(stride):
            args.append(tmp + j)
        return np.ravel(np.column_stack(args))


################################################################
# ParticleArrayExchange
################################################################w
cdef class ParticleArrayExchange:
    def __init__(self, int pa_index, ParticleArray pa, object comm):
        self.pa_index = pa_index
        self.pa = pa
        self.pa_wrapper = NNPSParticleArrayWrapper( pa )

        # unique data and message length tags for MPI communications
        name = pa.name
        self.msglength_tag_remote = sum( [ord(c) for c in name + '_msglength_remote'] )
        self.data_tag_remote = sum( [ord(c) for c in name + '_data_remote'] )

        self.msglength_tag_lb = sum( [ord(c) for c in name + '_msglength_lb'] )
        self.data_tag_lb = sum( [ord(c) for c in name + '_data_lb'] )

        # MPI comm rank and size
        self.comm = comm
        self.rank = comm.Get_rank()
        self.size = comm.Get_size()

        # num particles
        self.num_local = pa.get_number_of_particles()
        self.num_remote = 0
        self.num_global = 0
        self.num_ghost = 0

        # Particle Import/Export lists
        self.exportParticleGlobalids = UIntArray()
        self.exportParticleLocalids = UIntArray()
        self.exportParticleProcs = IntArray()
        self.numParticleExport = 0

        self.importParticleGlobalids = UIntArray()
        self.importParticleLocalids = UIntArray()
        self.importParticleProcs = IntArray()
        self.numParticleImport = 0

        # load balancing props are taken from the particle array
        lb_props = list(pa.get_lb_props())
        lb_props.sort()

        self.lb_props = lb_props
        self.nprops = len( lb_props )
        # If an lb_weight  constant is present in the array, this
        # is the weight to use for the load balancing for those particles.
        if 'lb_weight' in pa.constants:
            self.lb_weight = pa.constants.get('lb_weight')[0]
        else:
            self.lb_weight = 1.0

        # exchange flags
        self.lb_exchange = True
        self.remote_exchange = True

    def lb_exchange_data(self):
        """Share particle info after Zoltan_LB_Balance

        After an initial call to 'Zoltan_LB_Balance', the new size of
        the arrays should be (num_particles - numExport + numImport)
        to reflect particles that should be imported and exported.

        This function should be called after the load balancing lists
        a defined. The particles to be exported are removed and the
        arrays re-sized. MPI is then used to send and receive the data

        """
        _p = ProfileContext('PAExchange.lb_exchange_data')
        # data array
        cdef ParticleArray pa = self.pa

        # Export lists for the particles
        cdef UIntArray exportGlobalids = self.exportParticleGlobalids
        cdef UIntArray exportLocalids = self.exportParticleLocalids
        cdef IntArray exportProcs = self.exportParticleProcs
        cdef int numImport, numExport = self.numParticleExport

        # MPI communicator
        cdef object comm = self.comm

        # message tags
        cdef int dtag = self.data_tag_lb

        # create the zoltan communicator and get number of objects to import
        cdef ZComm zcomm = ZComm(comm, dtag, numExport, exportProcs.get_npy_array())
        numImport = zcomm.nreturn

        # extract particles to be exported
        cdef dict sendbufs = self.get_sendbufs( exportLocalids )

        # remove particles to be exported
        pa.remove_particles(exportLocalids, align=True)

        # old and new sizes
        cdef int current_size = self.num_local
        cdef int count = current_size - numExport
        cdef int newsize = count + numImport

        # resize the particle array to accomodate the receive.
        # remove_particles updates the size, so count is the
        # current size of the array.
        self.extend(count, newsize)

        # exchange data
        self.exchange_data(zcomm, sendbufs, count)

        # set all particle tags to local
        self.set_tag(count, newsize, Local)
        pa.align_particles()

        # update the number of particles
        self.num_local = newsize
        _p.stop()

    def remote_exchange_data(self):
        """Share particle info after computing remote particles.

        After a load balancing and the corresponding exhange of
        particles, additional particles are needed as remote
        (remote). Calls to 'compute_remote_particles' and
        'Zoltan_Invert_Lists' builds the lists required.

        The arrays must now be re-sized to (num_particles + numImport)
        to reflect the new particles that come in as remote particles.

        """
        _p = ProfileContext('PAExchange.remote_exchange_data')
        # data arrays
        cdef ParticleArray pa = self.pa

        # Export lists
        cdef UIntArray exportGlobalids = self.exportParticleGlobalids
        cdef UIntArray exportLocalids = self.exportParticleLocalids
        cdef IntArray exportProcs = self.exportParticleProcs
        cdef int numImport, numExport = self.numParticleExport

        # current number of particles
        cdef int current_size = self.num_local
        cdef int count, new_size

        # MPI communicator
        cdef object comm = self.comm

        # zoltan communicator
        cdef int dtag = self.data_tag_remote
        cdef ZComm zcomm = ZComm(comm, dtag, numExport, exportProcs.get_npy_array())
        numImport = zcomm.nreturn

        # old and new sizes
        count = current_size
        newsize = current_size + numImport

        # copy particles to be exported
        cdef dict sendbufs = self.get_sendbufs( exportLocalids )

        # update the size of the array
        self.extend(current_size, newsize)
        self.exchange_data(zcomm, sendbufs, count)

        # set tags for all received particles as Remote
        self.set_tag(count, newsize, Remote)
        pa.align_particles()

        # store the number of remote particles
        self.num_remote = newsize - current_size
        _p.stop()

    cdef exchange_data(self, ZComm zcomm, dict sendbufs, int count):
        cdef ParticleArray pa = self.pa
        cdef str prop
        cdef int prop_tag, stride, nbytes, i, nprops = self.nprops

        cdef np.ndarray prop_arr, sendbuf, recvbuf

        cdef list props = self.lb_props

        with profile_ctx('PAExchange.exchange_data'):
            for i in range(nprops):
                prop = props[i]

                prop_arr = pa.properties[prop].get_npy_array()
                nbytes = prop_arr.dtype.itemsize
                stride = pa.stride.get(prop, 1)

                # the send and receive buffers
                sendbuf = sendbufs[prop]
                recvbuf = prop_arr[count*stride:]

                # tag for this property
                prop_tag = i

                # set the nbytes and tag for the zoltan communicator
                zcomm.set_nbytes( nbytes*stride )
                zcomm.set_tag( prop_tag )

                # exchange the data
                zcomm.Comm_Do( sendbuf, recvbuf )

    def remove_remote_particles(self):
        self.num_local = self.pa.get_number_of_particles(real=True)
        cdef int num_local = self.num_local

        # resize the particle array
        self.pa.resize(num_local)
        self.pa.align_particles()

        # reset the number of remote particles
        self.num_remote = 0

    def align_particles(self):
        self.pa.align_particles()

    def set_tag(self, int start, int end, int value):
        """Reset the annoying tag value after particles are resized."""
        cdef int i
        cdef IntArray tag = self.pa_wrapper.tag

        for i in range(start, end):
            tag[i] = value

    def update_particle_gids(self):
        """Update the global indices.

        We call a utility function to get the new number of particles
        across the processors and then linearly assign indices to the
        objects.

        """
        cdef int num_global_objects, num_local_objects, _sum, i
        cdef np.ndarray[ndim=1, dtype=np.int32_t] num_objects_data

        # the global indices array
        cdef UIntArray gid = self.pa_wrapper.gid

        cdef object comm = self.comm
        cdef int rank = self.rank
        cdef int size = self.size

        num_objects_data = zoltan_utils.get_num_objects_per_proc(
            comm, self.num_local)

        num_local_objects = num_objects_data[ rank ]
        num_global_objects = np.sum( num_objects_data )

        _sum = np.sum( num_objects_data[:rank] )

        gid.resize( num_local_objects )
        for i in range( num_local_objects ):
            gid.data[i] = <ZOLTAN_ID_TYPE> ( _sum + i )

        # set the number of local and global objects
        self.num_global = num_global_objects

    def reset_lists(self):
        """Reset the particle lists"""
        self.numParticleExport = 0
        self.numParticleImport = 0

        self.exportParticleGlobalids.reset()
        self.exportParticleLocalids.reset()
        self.exportParticleProcs.reset()

        self.importParticleGlobalids.reset()
        self.importParticleLocalids.reset()
        self.importParticleProcs.reset()

    def extend(self, int currentsize, int newsize):
        if newsize <= currentsize:
            self.pa.resize(newsize)
        else:
            # We use extend so any non lb_prop props are reset.
            # resize does not set the values to any defaults.
            self.pa.extend(newsize - currentsize)

    def get_sendbufs(self, UIntArray exportIndices):
        cdef ParticleArray pa = self.pa
        cdef dict sendbufs = {}
        cdef str prop
        cdef int stride

        cdef np.ndarray indices = exportIndices.get_npy_array()

        strides = set(pa.stride.values())
        s_indices = {1:indices}
        for stride in strides:
            if stride > 1:
                s_indices[stride] = get_strided_indices(indices, stride)

        for prop in self.lb_props:
            prop_arr = pa.properties[prop].get_npy_array()
            stride = pa.stride.get(prop, 1)
            sendbufs[prop] = prop_arr[s_indices[stride]]

        return sendbufs

# #################################################################
# # ParallelManager extension classes
# #################################################################
cdef class ParallelManager:
    """Base class for all parallel managers."""
    def __init__(self, int dim, list particles, object comm,
                 double radius_scale=2.0,
                 int ghost_layers=2, domain=None,
                 bint update_cell_sizes=True):
        """Constructor.

        Parameters
        ----------

        dim : int
            Dimension

        particles : list
            list of particle arrays to be managed.

        comm : mpi4py.MPI.COMM, default
            MPI communicator for parallel invocations

        radius_scale : double, default (2)
            Optional kernel radius scale. Defaults to 2

        ghost_layers : int, default (2)
            Optional factor for computing bounding boxes for cells.

        domain : DomainManager, default (None)
            Optional limits for the domain

        """
        # number of arrays and a reference to the particle list
        self.narrays = len(particles)
        self.particles = particles

        # particle array exchange instances
        self.pa_exchanges = [ParticleArrayExchange(i, pa, comm) \
                             for i, pa in enumerate(particles)]

        # particle array wrappers
        self.pa_wrappers = [exchange.pa_wrapper for exchange in self.pa_exchanges]

        # number of local/global/remote particles
        self.num_local = [exchange.num_local for exchange in self.pa_exchanges]
        self.num_global = [0] * self.narrays
        self.num_remote = [0] * self.narrays

        # global indices used for load balancing
        self.cell_gid = UIntArray()

        # cell coordinate values used for load balancing
        self.cx = DoubleArray()
        self.cy = DoubleArray()
        self.cz = DoubleArray()

        # minimum and maximum arrays for MPI reduce operations. These
        # are used to find global bounds across processors.
        self.minx = np.zeros( shape=1, dtype=np.float64 )
        self.miny = np.zeros( shape=1, dtype=np.float64 )
        self.minz = np.zeros( shape=1, dtype=np.float64 )

        # global minimum values for x, y and z
        self.mx = 0.0; self.my = 0.0; self.mz = 0.0

        self.maxx = np.zeros( shape=1, dtype=np.float64 )
        self.maxy = np.zeros( shape=1, dtype=np.float64 )
        self.maxz = np.zeros( shape=1, dtype=np.float64 )
        self.maxh = np.zeros( shape=1, dtype=np.float64 )

        # global max values for x ,y, z & h
        self.Mx = 0.0; self.My = 0.0; self.Mz = 0.0; self.Mh = 0.0

        # MPI comm rank and size
        self.comm = comm
        self.rank = comm.Get_rank()
        self.size = comm.Get_size()

        self.in_parallel = True
        if self.size == 1: self.in_parallel = False

        # The cell_map dictionary, radius scale for binning and ghost
        # layers for remote neighbors.
        self.cell_map = {}
        self.cell_list = []

        # number of loca/remote cells
        self.ncells_local = 0
        self.ncells_remote = 0
        self.ncells_total = 0

        # radius scale and ghost layers
        self.radius_scale = radius_scale
        self.ghost_layers = ghost_layers

        # setup cell import/export lists
        self._setup_arrays()

        # flags to re-compute cell sizes
        self.initial_update = True
        self.update_cell_sizes = update_cell_sizes

        # update the particle global ids at startup
        self.update_particle_gids()
        self.local_bin()

        # load balancing frequency and counter
        self.lb_count = 0
        self.lb_freq = 1

        # array for global reduction of time steps
        self.dt_sendbuf = np.array( [1.0], dtype=np.float64 )

    def update_time_steps(self, double local_dt):
        """Peform a reduction to compute the globally stable time steps"""
        cdef np.ndarray dt_sendbuf = self.dt_sendbuf
        cdef np.ndarray dt_recvbuf = np.zeros_like(dt_sendbuf)

        comm = self.comm

        # set the local time step and peform the global reduction
        dt_sendbuf[0] = local_dt
        comm.Allreduce( sendbuf=dt_sendbuf, recvbuf=dt_recvbuf, op=mpi.MIN )

        return dt_recvbuf[0]

    cpdef compute_cell_size(self):
        """Compute the cell size for the binning.

        The cell size is chosen as the kernel radius scale times the
        maximum smoothing length in the local processor. For parallel
        runs, we would need to communicate the maximum 'h' on all
        processors to decide on the appropriate binning size.

        """
        # compute global bounds for the solution
        self._compute_bounds()
        cdef double hmax = self.Mh

        cell_size = self.radius_scale * hmax
        if cell_size < 1e-6:
            msg = """Cell size too small %g. Perhaps h = 0?
            Setting cell size to 1"""%(cell_size)
            print(msg)
            cell_size = 1.0
        self.cell_size = cell_size

    def update_cell_gids(self):
        """Update global indices for the cell_map dictionary.

        The objects to be partitioned in this class are the cells and
        we need to number them uniquely across processors. The
        numbering is done sequentially starting from the processor
        with rank 0 to the processor with rank size-1

        """
        # update the cell gids
        cdef PyZoltan pz = self.pz

        pz.num_local_objects = PyList_GET_SIZE( self.cell_list )
        pz._update_gid( self.cell_gid )

    def update_particle_gids(self):
        """Update individual particle global indices"""
        for i in range(self.narrays):
            pa_exchange = self.pa_exchanges[i]
            pa_exchange.update_particle_gids()

            self.num_local[i] = pa_exchange.num_local
            self.num_global[i] = pa_exchange.num_global

    def update(self):
        _p = ProfileContext("ParallelManager.update")
        cdef int lb_freq = self.lb_freq
        cdef int lb_count = self.lb_count

        lb_count += 1

        # remove remote particles from a previous step
        self.remove_remote_particles()

        if ( lb_count == lb_freq ):
            with profile_ctx('ParallelManager.update_partition'):
                self.update_partition()
            self.lb_count = 0
        else:
            with profile_ctx('ParallelManager.migrate_partition'):
                self.migrate_partition()
            self.lb_count = lb_count
        _p.stop()

    def update_partition(self):
        """Update the partition.

        This is the main entry point for the parallel manager. Given
        particles distributed across processors, we bin them, assign
        unique global indices for the cells and particles and
        subsequently perform the following steps:

        (a) Call a load balancing function to get import and export
            lists for the cell indices.

        for each particle array:

            (b) Create particle import/export lists using the cell
                import/export lists

            (c) Call ParticleArrayExchange's lb_exchange data with these particle
                import/export lists to effect the data movement for particles.

        (d) Update the local cell map after the data has been exchanged.

        now that we have the updated map, for each particle array:

            (e) Compute remote particle import/export lists from the cell map

            (f) Call ParticleArrayExchange's remote_exchange_data with these lists
                to effect the data movement.

        now that the data movement is done,

        (g) Update the local cell map to accommodate remote particles.

        Notes
        -----

        Although not intended to be a 'parallel' NNPS, the
        ParallelManager can be used for this purpose. I don't think
        this is advisable in variable smoothing length simulations
        where we should be using a very coarse grained partitioning to
        ensure safe local computations.

        For two step integrators commonly employed in SPH and with a
        sufficient number of ghost layers for remote particles, we
        should call 'update' only at the end of a time step. The idea
        of multiple ghost layers is to ensure that local particles
        have a sufficiently large halo region around them.

        """
        # update particle gids, bin particles and update cell gids
        #self.update_particle_gids()
        self.local_bin()
        self.update_cell_gids()

        if self.in_parallel:
            # use Zoltan to get the cell import/export lists
            with profile_ctx('ParallelManager.load_balance'):
                self.load_balance()

            # move the data to the new partitions
            if self.changes == 1:
                self.create_particle_lists()

                # exchange the data for each array
                for i in range(self.narrays):
                    self.pa_exchanges[i].lb_exchange_data()

                # update the local cell map after data movement
                self.update_local_data()

            # compute remote particles and exchange data
            self.compute_remote_particles()

            for i in range(self.narrays):
                self.pa_exchanges[i].remote_exchange_data()

            # update the local cell map to accommodate remote particles
            self.update_remote_data()

            # set the particle pids now that we have the partitions
            for i in range(self.narrays):
                pa = self.particles[i]
                pa.set_pid( self.rank )

    def migrate_partition(self):
        # migrate particles
        self.migrate_particles()

        # update local data
        self.update_local_data()

        # compute remote particles and exchange data
        self.compute_remote_particles()
        for i in range(self.narrays):
            self.pa_exchanges[i].remote_exchange_data()

        # update the local cell map to accomodate remotes
        self.update_remote_data()

        # set the particle pids
        for i in range(self.narrays):
            self.particles[i].set_pid( self.rank )

    def remove_remote_particles(self):
        """Remove remote particles"""
        cdef int narrays = self.narrays
        cdef ParticleArrayExchange pa_exchange

        with profile_ctx('ParallelManager.remove_remote_particles'):
            for i in range(narrays):
                pa_exchange = self.pa_exchanges[i]
                pa_exchange.remove_remote_particles()

                self.num_local[i] = pa_exchange.num_local
                self.num_remote[i] = pa_exchange.num_remote

    def local_bin(self):
        """Create the local cell map.

        Bin the particles by deleting any previous cells and
        re-computing the indexing structure. This corresponds to a
        local binning process that is called when each processor has
        a given list of particles to deal with.

        """
        cdef dict cell_map = self.cell_map
        cdef int num_particles
        cdef UIntArray indices

        # clear the indices for the cell_map.
        self.cell_map.clear()
        self.cell_list = []
        self.ncells_total = 0

        # compute the cell size
        if self.initial_update or self.update_cell_sizes:
            self.compute_cell_size()

        # # deal with ghosts
        # if self.is_periodic:
        #     # remove-ghost-particles
        #     self._remove_ghost_particles()

        #     # adjust local particles
        #     self._adjust_particles()

        #     # create new ghosts
        #     self._create_ghosts()

        # bin each particle array separately
        for i in range(self.narrays):
            num_particles = self.num_local[i]

            # bin the particles
            indices = arange_uint(num_particles)
            self._bin( i, indices )

        # local number of cells at this point are the total number of cells
        self.ncells_local = self.ncells_total

    cdef _bin(self, int pa_index, UIntArray indices):
        """Bin a given particle array with indices.

        Parameters
        ----------

        pa_index : int
            Index of the particle array corresponding to the particles list

        indices : UIntArray
            Subset of particles to bin

        """
        cdef NNPSParticleArrayWrapper pa_wrapper = self.pa_wrappers[ pa_index ]
        cdef DoubleArray x = pa_wrapper.x
        cdef DoubleArray y = pa_wrapper.y
        cdef DoubleArray z = pa_wrapper.z
        cdef UIntArray gid = pa_wrapper.gid

        cdef dict cell_map = self.cell_map
        cdef list cell_list = self.cell_list
        cdef double cell_size = self.cell_size

        cdef UIntArray lindices, gindices
        cdef size_t num_particles, indexi
        cdef ZOLTAN_ID_TYPE i

        cdef Cell cell
        cdef cIntPoint _cid
        cdef IntPoint cid = IntPoint()
        cdef cPoint pnt

        cdef int ncells_total = 0
        cdef int narrays = self.narrays
        cdef int layers = self.ghost_layers
        cdef double lb_weight = self.pa_exchanges[pa_index].lb_weight

        # now bin the particles
        num_particles = indices.length
        for indexi in range(num_particles):
            i = indices.data[indexi]

            pnt = cPoint_new( x.data[i], y.data[i], z.data[i] )
            _cid = find_cell_id( pnt, cell_size )

            cid = IntPoint_from_cIntPoint(_cid)

            if not cell_map.has_key( cid ):
                cell = Cell(cid=cid, cell_size=cell_size,
                            narrays=self.narrays, layers=layers)

                # store this cell in the cells list and cell map
                PyList_Append( cell_list, cell )
                cell_map[ cid ] = cell

                ncells_total = ncells_total + 1

            # add this particle to the list of indicies
            cell = cell_map[ cid ]

            lindices = cell.lindices[pa_index]
            gindices = cell.gindices[pa_index]

            lindices.append( i )
            gindices.append( gid.data[i] )

            # increment the size of the cells to reflect the added
            # particle. The size will be used to set the weights.
            cell.size += lb_weight

        # update the total number of cells
        self.ncells_total = self.ncells_total + ncells_total

    def update_local_data(self):
        """Update the cell map after load balance.

        After the load balancing step, each processor has a new set of
        local particles which need to be indexed. This new cell
        structure is then used to compute remote particles.

        """
        cdef ParticleArrayExchange pa_exchange
        cdef int num_particles, i
        cdef UIntArray indices

        # clear the local cell_map dict
        self.cell_map.clear()
        self.cell_list = []
        self.ncells_total = 0

        for i in range(self.narrays):
            pa_exchange = self.pa_exchanges[i]
            num_particles = pa_exchange.num_local

            # set the number of local and global particles
            self.num_local[i] = pa_exchange.num_local
            self.num_global[i] = pa_exchange.num_global

            # bin the particles
            indices = arange_uint( num_particles )
            self._bin(i, indices)

        self.ncells_local = self.ncells_total

    def update_remote_data(self):
        """Update the cell structure after sharing remote particles.

        After exchanging remote particles, we need to index the new
        remote particles for a possible neighbor query.

        """
        cdef int narrays = self.narrays
        cdef int num_local, num_remote, i
        cdef ParticleArrayExchange pa_exchange
        cdef UIntArray indices

        for i in range(narrays):
            pa_exchange = self.pa_exchanges[i]

            # get the number of local and remote particles
            num_local = pa_exchange.num_local
            num_remote = pa_exchange.num_remote

            # update the number of local/remote particles
            self.num_local[i] = num_local
            self.num_remote[i] = num_remote

            indices = arange_uint( num_local, num_local + num_remote )
            self._bin( i, indices )

        # compute the number of remote cells added
        self.ncells_remote = self.ncells_total - self.ncells_local

    def save_partition(self, fname, count=0):
        """Collect cell data from processors and save"""
        # get the global number of cells
        cdef cPoint centroid
        cdef int ncells_total = self.ncells_total
        cdef int ncells_local = self.ncells_local

        # cell centroid arrays
        cdef double[:] x = np.zeros( shape=ncells_total, dtype=np.float64 )
        cdef double[:] y = np.zeros( shape=ncells_total, dtype=np.float64 )
        cdef int[:] lid = np.zeros( shape=ncells_total, dtype=np.int32 )

        for i in range( ncells_total ):
            cell = self.cell_list[i]; centroid = cell.centroid
            x[i] = centroid.x; y[i] = centroid.y
            lid[i] = i

        cdef int[:] tag = np.zeros(shape=(ncells_total,), dtype=np.int32)
        tag[ncells_local:] = 1

        cdef double cell_size = self.cell_size

        # save the partition locally
        fname = fname + '/partition%03d.%d'%(count, self.rank)
        np.savez(
            fname, x=x, y=y, lid=lid, tag=tag, cell_size=cell_size,
            ncells_local=self.ncells_local, ncells_total=self.ncells_total
        )

    def set_lb_freq(self, int lb_freq):
        self.lb_freq = lb_freq

    def load_balance(self):
        raise NotImplementedError("ParallelManager::load_balance")

    def create_particle_lists(self, pa_index):
        raise NotImplementedError("ParallelManager::create_particle_lists")

    def compute_remote_particles(self, pa_index):
        raise NotImplementedError("ParallelManager::compute_remote_particles")

    #######################################################################
    # Private interface
    #######################################################################
    def set_data(self):
        """Compute the user defined data for use with Zoltan"""
        raise NotImplementedError("ZoltanParallelManager::_set_data should not be called!")

    def _setup_arrays(self):
        # Import/Export lists for cells
        self.exportCellGlobalids = UIntArray()
        self.exportCellLocalids = UIntArray()
        self.exportCellProcs = IntArray()

        self.importCellGlobalids = UIntArray()
        self.importCellLocalids = UIntArray()
        self.importCellProcs = IntArray()

        self.numCellImport = 0
        self.numCellExport = 0

    cdef _compute_bounds(self):
        """Compute the domain bounds for indexing."""

        cdef double mx, my, mz
        cdef double Mx, My, Mz, Mh

        cdef list pa_wrappers = self.pa_wrappers
        cdef NNPSParticleArrayWrapper pa
        cdef DoubleArray x, y, z, h

        # set some high and low values
        cdef double high = -1e20, low = 1e20
        mx = low; my = low; mz = low
        Mx = high; My = high; Mz = high; Mh = high

        # find the local min and max for all arrays on this proc
        for pa in pa_wrappers:
            x = pa.x; y = pa.y; z = pa.z; h = pa.h
            if x.length > 0:
                x.update_min_max(); y.update_min_max()
                z.update_min_max(); h.update_min_max()

                if x.minimum < mx: mx = x.minimum
                if x.maximum > Mx: Mx = x.maximum

                if y.minimum < my: my = y.minimum
                if y.maximum > My: My = y.maximum

                if z.minimum < mz: mz = z.minimum
                if z.maximum > Mz: Mz = z.maximum

                if h.maximum > Mh: Mh = h.maximum

        self.minx[0] = mx; self.miny[0] = my; self.minz[0] = mz
        self.maxx[0] = Mx; self.maxy[0] = My; self.maxz[0] = Mz
        self.maxh[0] = Mh

        # now compute global min and max if in parallel
        comm = self.comm
        if self.in_parallel:

            # revc buffers for all reduce
            _minx = np.zeros_like(self.minx)
            _miny = np.zeros_like(self.miny)
            _minz = np.zeros_like(self.minz)

            _maxx = np.zeros_like(self.maxx)
            _maxy = np.zeros_like(self.maxy)
            _maxz = np.zeros_like(self.maxz)
            _maxh = np.zeros_like(self.maxh)

            # global reduction for minimum values
            comm.Allreduce(sendbuf=self.minx, recvbuf=_minx, op=mpi.MIN)
            comm.Allreduce(sendbuf=self.miny, recvbuf=_miny, op=mpi.MIN)
            comm.Allreduce(sendbuf=self.minz, recvbuf=_minz, op=mpi.MIN)

            # global reduction for maximum values
            comm.Allreduce(sendbuf=self.maxx, recvbuf=_maxx, op=mpi.MAX)
            comm.Allreduce(sendbuf=self.maxy, recvbuf=_maxy, op=mpi.MAX)
            comm.Allreduce(sendbuf=self.maxz, recvbuf=_maxz, op=mpi.MAX)
            comm.Allreduce(sendbuf=self.maxh, recvbuf=_maxh, op=mpi.MAX)

        self.mx = _minx[0]; self.my = _miny[0]; self.mz = _minz[0]
        self.Mx = _maxx[0]; self.My = _maxy[0]; self.Mz = _maxz[0]
        self.Mh = _maxh[0]

    ######################################################################
    # Neighbor location routines
    ######################################################################
    cpdef get_nearest_particles(self, int src_index, int dst_index,
                                size_t d_idx, UIntArray nbrs):
        """Utility function to get near-neighbors for a particle.

        Parameters
        ----------

        src_index : int
            Index of the source particle array in the particles list

        dst_index : int
            Index of the destination particle array in the particles list

        d_idx : int (input)
            Particle for which neighbors are sought.

        nbrs : UIntArray (output)
            Neighbors for the requested particle are stored here.

        """
        cdef dict cell_map = self.cell_map
        cdef Cell cell

        cdef NNPSParticleArrayWrapper src = self.pa_wrappers[ src_index ]
        cdef NNPSParticleArrayWrapper dst = self.pa_wrappers[ dst_index ]

        # Source data arrays
        cdef DoubleArray s_x = src.x
        cdef DoubleArray s_y = src.y
        cdef DoubleArray s_z = src.z
        cdef DoubleArray s_h = src.h

        # Destination particle arrays
        cdef DoubleArray d_x = dst.x
        cdef DoubleArray d_y = dst.y
        cdef DoubleArray d_z = dst.z
        cdef DoubleArray d_h = dst.h

        cdef double radius_scale = self.radius_scale
        cdef double cell_size = self.cell_size
        cdef UIntArray lindices
        cdef size_t indexj
        cdef ZOLTAN_ID_TYPE j

        cdef cPoint xi = cPoint_new(d_x.data[d_idx], d_y.data[d_idx], d_z.data[d_idx])
        cdef cIntPoint _cid = find_cell_id( xi, cell_size )
        cdef IntPoint cid = IntPoint_from_cIntPoint( _cid )
        cdef IntPoint cellid = IntPoint(0, 0, 0)

        cdef cPoint xj
        cdef double xij

        cdef double hi, hj
        hi = radius_scale * d_h.data[d_idx]

        cdef int nnbrs = 0

        cdef int ix, iy, iz
        for ix in [cid.data.x -1, cid.data.x, cid.data.x + 1]:
            for iy in [cid.data.y - 1, cid.data.y, cid.data.y + 1]:
                for iz in [cid.data.z -1, cid.data.z, cid.data.z + 1]:
                    cellid.data.x = ix; cellid.data.y = iy
                    cellid.data.z = iz

                    if cell_map.has_key( cellid ):
                        cell = cell_map[ cellid ]
                        lindices = cell.lindices[src_index]
                        for indexj in range( lindices.length ):
                            j = lindices.data[indexj]

                            xj = cPoint_new( s_x.data[j], s_y.data[j], s_z.data[j] )
                            xij = cPoint_distance( xi, xj )

                            hj = radius_scale * s_h.data[j]

                            if ( (xij < hi) or (xij < hj) ):
                                if nnbrs == nbrs.length:
                                    nbrs.resize( nbrs.length + 50 )
                                    print('Neighbor search :: Extending the neighbor list to %d'%(nbrs.length))

                                nbrs.data[ nnbrs ] = j
                                nnbrs = nnbrs + 1

        # update the length for nbrs to indicate the number of neighbors
        nbrs.length = nnbrs

cdef class ZoltanParallelManager(ParallelManager):
    """Base class for Zoltan enabled parallel cell managers.

    To partition a list of arrays, we do an NNPS like box sort on all
    arrays to create a global spatial indexing structure. The cell_map
    are then used as 'objects' to be partitioned by Zoltan. The cell_map
    dictionary (cell-map) need not be unique across processors. We are
    responsible for assignning unique global ids for the cells.

    The Zoltan generated (cell) import/export lists are then used to
    construct particle import/export lists which are used to perform
    the data movement of the particles. A ParticleArrayExchange object
    (lb_exchange_data and remote_exchange_data) is used for each
    particle array in turn to effect this movement.

    """
    def __init__(self, int dim, list particles, object comm,
                 double radius_scale=2.0,
                 int ghost_layers=2, domain=None,
                 bint update_cell_sizes=True):
        """Constructor.

        Parameters
        ----------

        dim : int
            Dimension

        particles : list
            list of particle arrays to be managed.

        comm : mpi4py.MPI.COMM, default
            MPI communicator for parallel invocations

        radius_scale : double, default (2)
            Optional kernel radius scale. Defaults to 2

        ghost_layers : int, default (2)
            Optional factor for computing bounding boxes for cells.

        domain : DomainManager, default (None)
            Optional limits for the domain

        """
        super(ZoltanParallelManager, self).__init__(
            dim, particles, comm, radius_scale, ghost_layers, domain, update_cell_sizes)

        # Initialize the base PyZoltan class.
        self.pz = PyZoltan(comm)

    def create_particle_lists(self):
        """Create particle export/import lists

        For the cell based partitioner, Zoltan generated import and
        export lists apply to the cells. From this data, we can
        generate local export indices for the particles which must
        then be inverted to get the requisite import lists.

        We first construct the particle export lists in local arrays
        using the information from Zoltan_LB_Balance and subsequently
        call Zoltan_Invert_Lists to get the import lists. These arrays
        are then copied to the ParticleArrayExchange import/export
        lists.

        """
        # these are the Zoltan generated lists that correspond to cells
        cdef UIntArray exportCellLocalids = self.exportCellLocalids
        cdef UIntArray exportCellGlobalids = self.exportCellGlobalids
        cdef IntArray exportCellProcs = self.exportCellProcs
        cdef int numCellExport = self.numCellExport

        # the local cell map and cellids
        cdef int narrays = self.narrays
        cdef list cell_list = self.cell_list
        cdef ParticleArrayExchange pa_exchange

        cdef UIntArray exportLocalids, exportGlobalids
        cdef IntArray exportProcs

        cdef int indexi, i, j, export_proc, nindices, pa_index
        cdef UIntArray lindices, gindices
        cdef IntPoint cellid
        cdef Cell cell

        # initialize the particle export lists
        for pa_index in range(self.narrays):
            pa_exchange = self.pa_exchanges[pa_index]
            pa_exchange.reset_lists()

        # Iterate over the cells to be exported
        for indexi in range(numCellExport):
            i = <int>exportCellLocalids[indexi]            # local index for the cell to export
            cell = cell_list[i]                            # local cell to export
            export_proc = exportCellProcs.data[indexi]     # processor to export cell to

            # populate the export lists for each array in this cell
            for pa_index in range(self.narrays):
                # the ParticleArrayExchange object for this array
                pa_exchange = self.pa_exchanges[pa_index]

                exportLocalids = pa_exchange.exportParticleLocalids
                exportGlobalids = pa_exchange.exportParticleGlobalids
                exportProcs = pa_exchange.exportParticleProcs

                # local and global indices in this cell
                lindices = cell.lindices[pa_index]
                gindices = cell.gindices[pa_index]
                nindices = lindices.length

                for j in range( nindices ):
                    exportLocalids.append( lindices.data[j] )
                    exportGlobalids.append( gindices.data[j] )
                    exportProcs.append( export_proc )

        # save the number of particles
        for pa_index in range(narrays):
            pa_exchange = self.pa_exchanges[pa_index]
            pa_exchange.numParticleExport = pa_exchange.exportParticleProcs.length

    def compute_remote_particles(self):
        """Compute remote particles.

        Particles to be exported are determined by flagging individual
        cells and where they need to be shared to meet neighbor
        requirements.

        """
        # the PyZoltan object used to find intersections
        cdef PyZoltan pz = self.pz

        cdef Cell cell
        cdef cPoint boxmin, boxmax

        cdef object comm = self.comm
        cdef int rank = self.rank, narrays = self.narrays

        cdef list cell_list = self.cell_list
        cdef UIntArray lindices, gindices

        cdef np.ndarray nbrprocs
        cdef np.ndarray[ndim=1, dtype=np.int32_t] procs

        cdef int nbrproc, num_particles, indexi, pa_index
        cdef ZOLTAN_ID_TYPE i

        cdef int numprocs = 0

        cdef ParticleArrayExchange pa_exchange
        cdef UIntArray exportGlobalids, exportLocalids
        cdef IntArray exportProcs

        # reset the export lists
        for pa_index in range(narrays):
            pa_exchange = self.pa_exchanges[pa_index]
            pa_exchange.reset_lists()

        # Chaeck for each cell
        for cell in cell_list:

            # get the bounding box for this cell
            boxmin = cell.boxmin; boxmax = cell.boxmax

            numprocs = pz.Zoltan_Box_PP_Assign(
                boxmin.x, boxmin.y, boxmin.z,
                boxmax.x, boxmax.y, boxmax.z)

            # the array of processors that this box intersects with
            procs = pz.procs

            # array of neighboring processors
            nbrprocs = procs[np.where( (procs != -1) * (procs != rank) )[0]]

            if nbrprocs.size > 0:
                cell.is_boundary = True

                # store the neighboring processors for this cell
                cell.nbrprocs.resize( nbrprocs.size )
                cell.nbrprocs.set_data( nbrprocs )

                # populate the particle export lists for each array
                for pa_index in range(narrays):
                    pa_exchange = self.pa_exchanges[pa_index]

                    exportGlobalids = pa_exchange.exportParticleGlobalids
                    exportLocalids = pa_exchange.exportParticleLocalids
                    exportProcs = pa_exchange.exportParticleProcs

                    # local and global indices in this cell
                    lindices = cell.lindices[pa_index]
                    gindices = cell.gindices[pa_index]
                    num_particles = lindices.length

                    for nbrproc in nbrprocs:
                        for indexi in range( num_particles ):
                            i = lindices.data[indexi]

                            exportLocalids.append( i )
                            exportGlobalids.append( gindices.data[indexi] )
                            exportProcs.append( nbrproc )

        # set the numParticleExport for each array
        for pa_index in range(narrays):
            pa_exchange = self.pa_exchanges[pa_index]
            pa_exchange.numParticleExport = pa_exchange.exportParticleProcs.length

    def load_balance(self):
        """Use Zoltan to generate import/export lists for the cells.

        For the Zoltan interface, we require to populate a user
        defined struct with appropriate data for Zoltan to deal
        with. For the geometric based partitioners for example, we
        require the unique cell global indices and arrays for the cell
        centroid coordinates. Computation of this data must be done
        prior to calling 'pz.Zoltan_LB_Balance' through a call to
        'set_data'

        """
        cdef PyZoltan pz = self.pz

        # set the data which will be used by the Zoltan wrapper
        self.set_data()

        # call Zoltan to get the cell import/export lists
        self.changes = pz.Zoltan_LB_Balance()

        # copy the Zoltan export lists to the cell lists
        self.numCellExport = pz.numExport
        self.exportCellGlobalids.resize( pz.numExport )
        self.exportCellGlobalids.copy_subset( pz.exportGlobalids )

        self.exportCellLocalids.resize( pz.numExport )
        self.exportCellLocalids.copy_subset( pz.exportLocalids )

        self.exportCellProcs.resize( pz.numExport )
        self.exportCellProcs.copy_subset( pz.exportProcs )

        # copy the Zoltan import lists to the cell lists
        self.numCellImport = pz.numImport
        self.importCellGlobalids.resize( pz.numImport )
        self.importCellGlobalids.copy_subset( pz.importGlobalids )

        self.importCellLocalids.resize( pz.numImport )
        self.importCellLocalids.copy_subset( pz.importLocalids )

        self.importCellProcs.resize( pz.numImport )
        self.importCellProcs.copy_subset( pz.importProcs )

    def migrate_particles(self):
        raise NotImplementedError("ZoltanParallelManager::migrate_particles called")

cdef class ZoltanParallelManagerGeometric(ZoltanParallelManager):
    """Zoltan enabled parallel manager for use with geometric load balancing.

    Use this class for the Zoltan RCB, RIB and HSFC load balancing
    algorithms.

    """
    def __init__(self, int dim, list particles, object comm,
                 double radius_scale=2.0,
                 int ghost_layers=2, domain=None,
                 str lb_method='RCB',
                 str keep_cuts="1",
                 str obj_weight_dim="1",
                 bint update_cell_sizes=True
                 ):

        # initialize the base class
        super(ZoltanParallelManagerGeometric, self).__init__(
            dim, particles, comm, radius_scale, ghost_layers, domain,
            update_cell_sizes)

        # concrete implementation of a PyZoltan class
        self.pz = ZoltanGeometricPartitioner(
            dim, self.comm, self.cx, self.cy, self.cz, self.cell_gid,
            obj_weight_dim=obj_weight_dim,
            keep_cuts=keep_cuts,
            lb_method=lb_method)

    def set_zoltan_rcb_lock_directions(self):
        """Set the Zoltan flag to fix the directions of the RCB cuts"""
        self.pz.set_rcb_lock_directions('1')

    def set_zoltan_rcb_reuse(self):
        """Use previous cuts as guesses for the current RCB cut"""
        self.pz.set_rcb_reuse('1')

    def set_zoltan_rcb_rectilinear_blocks(self):
        """Flag to control the shape of the RCB regions """
        self.pz.set_rcb_rectilinear_blocks('1')

    def set_zoltan_rcb_directions(self, value):
        """Option to group the cuts along a given direction

        Legal values (refer to the Zoltan User Guide):

        '0' = don't order cuts;
        '1' = xyz
        '2' = xzy
        '3' = yzx
        '4' = yxz
        '5' = zxy
        '6' = zyx

        """
        self.pz.set_rcb_directions(value)

    def set_data(self):
        """Set the user defined particle data structure for Zoltan.

        For the geometric based partitioners, Zoltan requires as input
        the number of local/global objects, unique global indices for
        the local objects and coordinate values for the local
        objects.

        """
        cdef ZoltanGeometricPartitioner pz = self.pz
        cdef list cell_list = self.cell_list

        cdef int num_local_objects = pz.num_local_objects
        cdef double num_global_objects1 = 1.0/pz.num_global_objects
        cdef double weight

        cdef DoubleArray x = self.cx
        cdef DoubleArray y = self.cy
        cdef DoubleArray z = self.cz

        # the weights array for PyZoltan
        cdef DoubleArray weights = pz.weights

        cdef int i
        cdef Cell cell
        cdef cPoint centroid

        # resize the coordinate and PyZoltan weight arrays
        x.resize( num_local_objects )
        y.resize( num_local_objects )
        z.resize( num_local_objects )

        weights.resize( num_local_objects )

        # populate the arrays
        for i in range( num_local_objects ):
            cell = cell_list[ i ]
            centroid = cell.centroid

            # weights are defined as cell.size/num_total
            weights.data[i] = num_global_objects1 * cell.size

            x.data[i] = centroid.x
            y.data[i] = centroid.y
            z.data[i] = centroid.z

    def migrate_particles(self):
        """Update an existing partition"""
        cdef PyZoltan pz = self.pz
        cdef int pa_index, narrays=self.narrays
        cdef double xi, yi, zi

        cdef NNPSParticleArrayWrapper pa_wrapper
        cdef ParticleArrayExchange pa_exchange

        cdef UIntArray exportGlobalids, exportLocalids
        cdef IntArray exportParticleProcs

        cdef DoubleArray x, y, z
        cdef UIntArray gid

        cdef ZOLTAN_ID_TYPE local_id, global_id

        cdef int num_particles, i, proc
        cdef int rank = self.rank

        # iterate over all arrays
        for pa_index in range(narrays):
            pa_exchange = self.pa_exchanges[ pa_index ]
            pa_wrapper = self.pa_wrappers[ pa_index ]

            # reset the export lists
            pa_exchange.reset_lists()
            exportGlobalids = pa_exchange.exportParticleGlobalids
            exportLocalids = pa_exchange.exportParticleLocalids
            exportProcs = pa_exchange.exportParticleProcs

            # particle arrays
            x = pa_wrapper.x; y = pa_wrapper.y; z = pa_wrapper.z
            gid = pa_wrapper.gid

            num_particles = self.num_local[pa_index]
            for i in range(num_particles):
                xi = x.data[i]; yi = y.data[i]; zi = z.data[i]

                # find the processor to which this point belongs
                proc = pz.Zoltan_Point_PP_Assign(xi, yi, zi)
                if proc != rank:
                    exportGlobalids.append( gid.data[i] )
                    exportLocalids.append( <ZOLTAN_ID_TYPE>i )
                    exportProcs.append(proc)

            # exchange the data given the export lists
            pa_exchange.numParticleExport = exportProcs.length
            pa_exchange.lb_exchange_data()



"""
Generalized Transport Velocity Formulation
##########################################

Some notes on the paper,

- In the viscosity term of equation (17) a factor of '2' is missing.
- A negative sign is missing from equation (22) i.e, either put a negative
  sign in equation (22) or at the integrator step equation(25).
- The Solid Mechanics Equations are not tested.

References
-----------

    .. [ZhangHuAdams2017] Chi Zhang, Xiangyu Y. Hu, Nikolaus A. Adams
       "A generalized transport-velocity formulation for smoothed particle
       hydrodynamics", Journal of Computational Physics 237 (2017),
       pp. 216--232.
"""


from compyle.api import declare
from pysph.sph.equation import Equation
from pysph.base.utils import get_particle_array
from pysph.sph.integrator import Integrator
from pysph.sph.integrator_step import IntegratorStep
from pysph.sph.equation import Group, MultiStageEquations
from pysph.sph.scheme import Scheme
from pysph.sph.wc.linalg import mat_vec_mult, mat_mult


def get_particle_array_gtvf(constants=None, **props):
    gtvf_props = [
        'uhat', 'vhat', 'what', 'rho0', 'rhodiv', 'p0', 'auhat', 'avhat',
        'awhat', 'arho', 'arho0'
    ]

    pa = get_particle_array(
        constants=constants, additional_props=gtvf_props, **props
    )
    pa.add_property('gradvhat', stride=9)
    pa.add_property('sigma', stride=9)
    pa.add_property('asigma', stride=9)

    pa.set_output_arrays([
        'x', 'y', 'z', 'u', 'v', 'w', 'rho', 'p', 'h', 'm', 'au', 'av', 'aw',
        'pid', 'gid', 'tag'
    ])
    return pa


class GTVFIntegrator(Integrator):
    def one_timestep(self, t, dt):
        self.stage1()
        self.do_post_stage(dt, 1)

        self.compute_accelerations(0, update_nnps=False)

        self.stage2()
        # We update domain here alone as positions only change here.
        self.update_domain()
        self.do_post_stage(dt, 2)

        self.compute_accelerations(1)

        self.stage3()
        self.do_post_stage(dt, 3)


class GTVFStep(IntegratorStep):
    def stage1(self, d_idx, d_u, d_v, d_w, d_au, d_av, d_aw, d_uhat, d_vhat,
               d_what, d_auhat, d_avhat, d_awhat, dt):
        dtb2 = 0.5*dt
        d_u[d_idx] += dtb2*d_au[d_idx]
        d_v[d_idx] += dtb2*d_av[d_idx]
        d_w[d_idx] += dtb2*d_aw[d_idx]

        d_uhat[d_idx] = d_u[d_idx] + dtb2*d_auhat[d_idx]
        d_vhat[d_idx] = d_v[d_idx] + dtb2*d_avhat[d_idx]
        d_what[d_idx] = d_w[d_idx] + dtb2*d_awhat[d_idx]

    def stage2(self, d_idx, d_uhat, d_vhat, d_what, d_x, d_y, d_z, d_rho,
               d_arho, d_sigma, d_asigma, dt):
        d_rho[d_idx] += dt*d_arho[d_idx]

        i = declare('int')
        for i in range(9):
            d_sigma[d_idx*9 + i] += dt * d_asigma[d_idx*9 + i]

        d_x[d_idx] += dt*d_uhat[d_idx]
        d_y[d_idx] += dt*d_vhat[d_idx]
        d_z[d_idx] += dt*d_what[d_idx]

    def stage3(self, d_idx, d_u, d_v, d_w, d_au, d_av, d_aw, dt):
        dtb2 = 0.5*dt
        d_u[d_idx] += dtb2*d_au[d_idx]
        d_v[d_idx] += dtb2*d_av[d_idx]
        d_w[d_idx] += dtb2*d_aw[d_idx]


class ContinuityEquationGTVF(Equation):
    r"""**Evolution of density**

    From [ZhangHuAdams2017], equation (12),

    .. math::
            \frac{\tilde{d} \rho_i}{dt} = \rho_i \sum_j \frac{m_j}{\rho_j}
            \nabla W_{ij} \cdot \tilde{\boldsymbol{v}}_{ij}
    """
    def initialize(self, d_arho, d_idx):
        d_arho[d_idx] = 0.0

    def loop(self, d_idx, s_idx, s_m, d_rho, s_rho, d_uhat, d_vhat, d_what,
             s_uhat, s_vhat, s_what, d_arho, DWIJ):
        uhatij = d_uhat[d_idx] - s_uhat[s_idx]
        vhatij = d_vhat[d_idx] - s_vhat[s_idx]
        whatij = d_what[d_idx] - s_what[s_idx]

        udotdij = DWIJ[0]*uhatij + DWIJ[1]*vhatij + DWIJ[2]*whatij
        fac = d_rho[d_idx] * s_m[s_idx] / s_rho[s_idx]
        d_arho[d_idx] += fac * udotdij


class CorrectDensity(Equation):
    r"""**Density correction**

    From [ZhangHuAdams2017], equation (13),

    .. math::
            \rho_i = \frac{\sum_j m_j W_{ij}}
            {\min(1, \sum_j \frac{m_j}{\rho_j^{*}} W_{ij})}

    where,

    .. math::
            \rho_j^{*} = \text{density before this correction is applied.}
    """
    def initialize(self, d_idx, d_rho, d_rho0, d_rhodiv):
        d_rho0[d_idx] = d_rho[d_idx]
        d_rho[d_idx] = 0.0
        d_rhodiv[d_idx] = 0.0

    def loop(self, d_idx, s_idx, d_rho, d_rhodiv, s_m, WIJ, s_rho0):
        d_rho[d_idx] += s_m[s_idx]*WIJ
        d_rhodiv[d_idx] += s_m[s_idx]*WIJ/s_rho0[s_idx]

    def post_loop(self, d_idx, d_rho, d_rhodiv):
        d_rho[d_idx] = d_rho[d_idx] / min(1, d_rhodiv[d_idx])


class MomentumEquationPressureGradient(Equation):
    r"""**Momentum Equation**

    From [ZhangHuAdams2017], equation (17),

    .. math::
            \frac{\tilde{d} \boldsymbol{v}_i}{dt} = - \sum_j m_j \nabla W_{ij}
            \cdot \left[\left(\frac{p_i}{\rho_i^2} + \frac{p_j}{\rho_j^2}
            \right)\textbf{I} - \left(\frac{\boldsymbol{A_i}}{\rho_i^2} +
            \frac{\boldsymbol{A_j}}{\rho_j^2} \right)\right] + \sum_j
            \frac{\eta_{ij}\boldsymbol{v}_{ij}}{\rho_i \rho_j r_{ij}}
            \nabla W_{ij} \cdot \boldsymbol{x}_{ij}

    where,

    .. math::
           \boldsymbol{A_{i/j}} = \rho_{i/j} \boldsymbol{v}_{i/j} \otimes
           (\tilde{\boldsymbol{v}}_{i/j} - \boldsymbol{v}_{i/j})

    .. math::
           \eta_{ij} = \frac{2\eta_i \eta_j}{\eta_i + \eta_j}
    .. math::
           \eta_{i/j} = \rho_{i/j} \nu

    for solids, replace :math:`\boldsymbol{A}_{i/j}` with
    :math:`\boldsymbol{\sigma}'_{i/j}`.

    The rate of change of transport velocity is given by,

    .. math::
            (\frac{d\boldsymbol{v}_i}{dt})_c = -p_i^0 \sum_j \frac{m_j}
            {\rho_i^2} \nabla \tilde{W}_{ij}

    where,

    .. math::
            \tilde{W}_{ij} = W(\boldsymbol{x}_ij, \tilde{0.5 h_{ij}})
    .. math::
            p_i^0 = \min(10|p_i|, p_{ref})

    Notes:

    A negative sign in :math:`(\frac{d\boldsymbol{v}_i}{dt})_c` is
    missing in the paper [ZhangHuAdams2017].
    """

    def __init__(self, dest, sources, pref, gx=0.0, gy=0.0, gz=0.0):
        r"""
        Parameters
        ----------
        pref : float
            reference pressure
        gx : float
            body force per unit mass along the x-axis
        gy : float
            body force per unit mass along the y-axis
        gz : float
            body force per unit mass along the z-axis
        """

        self.pref = pref
        self.gx = gx
        self.gy = gy
        self.gz = gz
        super(MomentumEquationPressureGradient, self).__init__(dest, sources)

    def initialize(self, d_idx, d_au, d_av, d_aw, d_auhat, d_avhat, d_awhat,
                   d_p0, d_p):
        d_au[d_idx] = self.gx
        d_av[d_idx] = self.gy
        d_aw[d_idx] = self.gz

        d_auhat[d_idx] = 0.0
        d_avhat[d_idx] = 0.0
        d_awhat[d_idx] = 0.0

        d_p0[d_idx] = min(10*abs(d_p[d_idx]), self.pref)

    def loop(self, d_rho, s_rho, d_idx, s_idx, d_p, s_p, s_m, d_au, d_av,
             d_aw, DWIJ, d_p0, d_auhat, d_avhat, d_awhat, XIJ, RIJ, SPH_KERNEL,
             HIJ):
        rhoi2 = d_rho[d_idx] * d_rho[d_idx]
        rhoj2 = s_rho[s_idx] * s_rho[s_idx]

        pij = d_p[d_idx]/rhoi2 + s_p[s_idx]/rhoj2

        tmp = -s_m[s_idx] * pij

        d_au[d_idx] += tmp * DWIJ[0]
        d_av[d_idx] += tmp * DWIJ[1]
        d_aw[d_idx] += tmp * DWIJ[2]

        tmp = -d_p0[d_idx] * s_m[s_idx]/rhoi2

        dwijhat = declare('matrix(3)')
        SPH_KERNEL.gradient(XIJ, RIJ, 0.5*HIJ, dwijhat)

        d_auhat[d_idx] += tmp * dwijhat[0]
        d_avhat[d_idx] += tmp * dwijhat[1]
        d_awhat[d_idx] += tmp * dwijhat[2]


class MomentumEquationViscosity(Equation):
    r"""**Momentum equation Artificial stress for solids**

    See the class MomentumEquationPressureGradient for details.

    Notes:

    A factor of '2' is missing in the viscosity equation given by
    [ZhangHuAdams2017].
    """
    def __init__(self, dest, sources, nu):
        r"""
        Parameters
        ----------
        nu : float
            viscosity of the fluid.
        """

        self.nu = nu
        super(MomentumEquationViscosity, self).__init__(dest, sources)

    def loop(self, d_idx, s_idx, d_rho, s_rho, s_m, d_au,
             d_av, d_aw, VIJ, R2IJ, EPS, DWIJ, XIJ):
        etai = self.nu * d_rho[d_idx]
        etaj = self.nu * s_rho[s_idx]

        etaij = 4 * (etai * etaj)/(etai + etaj)

        xdotdij = DWIJ[0]*XIJ[0] + DWIJ[1]*XIJ[1] + DWIJ[2]*XIJ[2]

        tmp = s_m[s_idx]/(d_rho[d_idx] * s_rho[s_idx])
        fac = tmp * etaij * xdotdij/(R2IJ + EPS)

        d_au[d_idx] += fac * VIJ[0]
        d_av[d_idx] += fac * VIJ[1]
        d_aw[d_idx] += fac * VIJ[2]


class MomentumEquationArtificialStress(Equation):
    r"""**Momentum equation Artificial stress for solids**

    See the class MomentumEquationPressureGradient for details.
    """
    def __init__(self, dest, sources, dim):
        r"""
        Parameters
        ----------
        dim : int
            Dimensionality of the problem.
        """
        self.dim = dim
        super(MomentumEquationArtificialStress, self).__init__(dest, sources)

    def _get_helpers_(self):
        return [mat_vec_mult]

    def loop(self, d_idx, s_idx, d_rho, s_rho, d_u, d_v, d_w, d_uhat, d_vhat,
             d_what, s_u, s_v, s_w, s_uhat, s_vhat, s_what, d_au, d_av, d_aw,
             s_m, DWIJ):
        rhoi = d_rho[d_idx]
        rhoj = s_rho[s_idx]

        i, j = declare('int', 2)
        ui, uj, uidif, ujdif, res = declare('matrix(3)', 5)
        Aij = declare('matrix(9)')

        for i in range(3):
            res[i] = 0.0
            for j in range(3):
                Aij[3*i + j] = 0.0

        ui[0] = d_u[d_idx]
        ui[1] = d_v[d_idx]
        ui[2] = d_w[d_idx]

        uj[0] = s_u[s_idx]
        uj[1] = s_v[s_idx]
        uj[2] = s_w[s_idx]

        uidif[0] = d_uhat[d_idx] - d_u[d_idx]
        uidif[1] = d_vhat[d_idx] - d_v[d_idx]
        uidif[2] = d_what[d_idx] - d_w[d_idx]

        ujdif[0] = s_uhat[s_idx] - s_u[s_idx]
        ujdif[1] = s_vhat[s_idx] - s_v[s_idx]
        ujdif[2] = s_what[s_idx] - s_w[s_idx]

        for i in range(3):
            for j in range(3):
                Aij[3*i + j] = (ui[i]*uidif[j] / rhoi + uj[i]*ujdif[j] / rhoj)

        mat_vec_mult(Aij, DWIJ, 3, res)

        d_au[d_idx] += s_m[s_idx] * res[0]
        d_av[d_idx] += s_m[s_idx] * res[1]
        d_aw[d_idx] += s_m[s_idx] * res[2]


class VelocityGradient(Equation):
    r"""**Gradient of velocity vector**

    .. math::
            (\nabla \otimes \tilde{\boldsymbol{v}})_i = \sum_j \frac{m_j}
            {\rho_j} \tilde{\boldsymbol{v}}_{ij} \otimes \nabla W_{ij}
    """
    def __init__(self, dest, sources, dim):
        r"""
        Parameters
        ----------
        dim : int
            Dimensionality of the problem.
        """
        self.dim = dim
        super(VelocityGradient, self).__init__(dest, sources)

    def initialize(self, d_idx, d_gradvhat):
        for i in range(9):
            d_gradvhat[9*d_idx + i] = 0.0

    def loop(self, s_idx, d_idx, s_m, d_uhat, d_vhat, d_what, s_uhat, s_vhat,
             s_what, s_rho, d_gradvhat, DWIJ):
        i, j = declare('int', 2)
        uhatij = declare('matrix(3)')
        Vj = s_m[s_idx]/s_rho[s_idx]

        uhatij[0] = d_uhat[d_idx] - s_uhat[s_idx]
        uhatij[1] = d_vhat[d_idx] - s_vhat[s_idx]
        uhatij[2] = d_what[d_idx] - s_what[s_idx]

        for i in range(3):
            for j in range(3):
                d_gradvhat[d_idx*9 + 3*i + j] += Vj * uhatij[i] * DWIJ[j]


class DeviatoricStressRate(Equation):
    r"""**Stress rate for solids**

    From [ZhangHuAdams2017], equation (5),

    .. math::
            \frac{d \boldsymbol{\sigma}'}{dt} = 2 G (\boldsymbol{\epsilon}
            - \frac{1}{3} \text{Tr}(\boldsymbol{\epsilon})\textbf{I}) +
            \boldsymbol{\sigma}' \cdot \boldsymbol{\Omega}^{T} +
            \boldsymbol{\Omega} \cdot \boldsymbol{\sigma}'

    where,

    .. math::
           \boldsymbol{\Omega_{i/j}} = \frac{1}{2}
           \left(\nabla \otimes \boldsymbol{v}_{i/j} -
           (\nabla \otimes \boldsymbol{v}_{i/j})^{T}\right)

    .. math::
           \boldsymbol{\epsilon_{i/j}} = \frac{1}{2}
           \left(\nabla \otimes \boldsymbol{v}_{i/j} +
           (\nabla \otimes \boldsymbol{v}_{i/j})^{T}\right)

    see the class VelocityGradient for :math:`\nabla \otimes \boldsymbol{v}_i`
   """
    def __init__(self, dest, sources, dim, G):
        r"""
        Parameters
        ----------
        dim : int
            Dimensionality of the problem.
        G : float
            value of shear modulus
        """
        self.G = G
        self.dim = dim
        super(DeviatoricStressRate, self).__init__(dest, sources)

    def _get_helpers_(self):
        return [mat_vec_mult, mat_mult]

    def initialize(self, d_idx, d_sigma, d_asigma, d_gradvhat):
        i, j, ind = declare('int', 3)
        eps, omega, omegaT, sigmai, dvi = declare('matrix(9)', 5)

        G = self.G

        for i in range(9):
            sigmai[i] = d_sigma[d_idx*9 + i]
            dvi[i] = d_gradvhat[d_idx*9 + i]
            d_asigma[d_idx*9 + i] = 0.0

        eps_trace = 0.0
        for i in range(3):
            for j in range(3):
                eps[3*i + j] = 0.5*(dvi[3*i + j] + dvi[3*j + i])
                omega[3*i + j] = 0.5*(dvi[3*i + j] - dvi[3*j + i])
                if i == j:
                    eps_trace += eps[3*i + j]

        for i in range(3):
            for j in range(3):
                omegaT[3*j + i] = omega[3*i + j]

        smo, oms = declare('matrix(9)', 2)
        mat_mult(sigmai, omegaT, 3, smo)
        mat_mult(omega, sigmai, 3, oms)

        for i in range(3):
            for j in range(3):
                ind = 3*i + j
                d_asigma[d_idx*9 + ind] = 2*G * eps[ind] + smo[ind] + oms[ind]
                if i == j:
                    d_asigma[d_idx*9 + ind] += -2*G * eps_trace/3.0


class MomentumEquationArtificialStressSolid(Equation):
    r"""**Momentum equation Artificial stress for solids**

    See the class MomentumEquationPressureGradient for details.
    """
    def __init__(self, dest, sources, dim):
        r"""
        Parameters
        ----------
        dim : int
            Dimensionality of the problem.
        """
        self.dim = dim
        super(MomentumEquationArtificialStressSolid, self).__init__(dest,
                                                                    sources)

    def _get_helpers_(self):
        return [mat_vec_mult]

    def loop(self, d_idx, s_idx, d_sigma, s_sigma, d_au, d_av, d_aw, s_m,
             DWIJ):
        i = declare('int')
        sigmaij = declare('matrix(9)')
        res = declare('matrix(3)')

        for i in range(9):
            sigmaij[i] = d_sigma[d_idx*9 + i] + s_sigma[s_idx*9 + i]

        mat_vec_mult(sigmaij, DWIJ, 3, res)

        d_au[d_idx] += s_m[s_idx] * res[0]
        d_av[d_idx] += s_m[s_idx] * res[1]
        d_aw[d_idx] += s_m[s_idx] * res[2]


class GTVFScheme(Scheme):
    def __init__(self, fluids, solids, dim, rho0, c0, nu, h0, pref,
                 gx=0.0, gy=0.0, gz=0.0, b=1.0, alpha=0.0):
        r"""Parameters
        ----------

        fluids: list
            List of names of fluid particle arrays.
        solids: list
            List of names of solid particle arrays.
        dim: int
            Dimensionality of the problem.
        rho0: float
            Reference density.
        c0: float
            Reference speed of sound.
        nu: float
            Real viscosity of the fluid.
        h0: float
            Reference smoothing length.
        pref: float
            reference pressure for rate of change of transport velocity.
        gx: float
            Body force acceleration components in x direction.
        gy: float
            Body force acceleration components in y direction.
        gz: float
            Body force acceleration components in z direction.
        b: float
            constant for the equation of state.
        """

        self.fluids = fluids
        self.solids = solids
        self.dim = dim
        self.rho0 = rho0
        self.c0 = c0
        self.nu = nu
        self.h0 = h0
        self.pref = pref
        self.gx = gx
        self.gy = gy
        self.gz = gz
        self.b = b
        self.alpha = alpha
        self.solver = None

    def configure_solver(self, kernel=None, integrator_cls=None,
                         extra_steppers=None, **kw):
        """Configure the solver to be generated.

        Parameters
        ----------

        kernel : Kernel instance.
            Kernel to use, if none is passed a default one is used.
        integrator_cls : pysph.sph.integrator.Integrator
            Integrator class to use, use sensible default if none is
            passed.
        extra_steppers : dict
            Additional integration stepper instances as a dict.
        **kw : extra arguments
            Any additional keyword args are passed to the solver instance.
        """
        from pysph.base.kernels import WendlandQuintic
        if kernel is None:
            kernel = WendlandQuintic(dim=self.dim)
        steppers = {}
        if extra_steppers is not None:
            steppers.update(extra_steppers)

        step_cls = GTVFStep
        for fluid in self.fluids:
            if fluid not in steppers:
                steppers[fluid] = step_cls()

        if integrator_cls is not None:
            cls = integrator_cls
            print("Warning: GTVF Integrator is not being used.")
        else:
            cls = GTVFIntegrator
        integrator = cls(**steppers)

        from pysph.solver.solver import Solver
        self.solver = Solver(
            dim=self.dim, integrator=integrator, kernel=kernel, **kw
        )

    def get_equations(self):
        from pysph.sph.wc.transport_velocity import (
            StateEquation, SetWallVelocity, SolidWallPressureBC,
            VolumeSummation, SolidWallNoSlipBC,
            MomentumEquationArtificialViscosity, ContinuitySolid
        )
        all = self.fluids + self.solids

        stage1 = []
        if self.solids:
            eq0 = []
            for solid in self.solids:
                eq0.append(SetWallVelocity(dest=solid, sources=self.fluids))
            stage1.append(Group(equations=eq0, real=False))

        eq1 = []
        for fluid in self.fluids:
            eq1.append(ContinuityEquationGTVF(dest=fluid, sources=self.fluids))
            if self.solids:
                eq1.append(
                    ContinuitySolid(dest=fluid, sources=self.solids)
                )
        stage1.append(Group(equations=eq1, real=False))

        eq2, stage2 = [], []
        for fluid in self.fluids:
            eq2.append(CorrectDensity(dest=fluid, sources=all))
        stage2.append(Group(equations=eq2, real=False))

        eq3 = []
        for fluid in self.fluids:
            eq3.append(
                StateEquation(dest=fluid, sources=None, p0=self.pref,
                              rho0=self.rho0, b=1.0)
            )
        stage2.append(Group(equations=eq3, real=False))

        g2_s = []
        for solid in self.solids:
            g2_s.append(VolumeSummation(dest=solid, sources=all))
            g2_s.append(SolidWallPressureBC(
                dest=solid, sources=self.fluids, b=1.0, rho0=self.rho0,
                p0=self.pref, gx=self.gx, gy=self.gy, gz=self.gz
            ))
        if g2_s:
            stage2.append(Group(equations=g2_s, real=False))

        eq4 = []
        for fluid in self.fluids:
            eq4.append(
                MomentumEquationPressureGradient(
                    dest=fluid, sources=all, pref=self.pref,
                    gx=self.gx, gy=self.gy, gz=self.gz
                ))
            if self.alpha > 0.0:
                eq4.append(
                    MomentumEquationArtificialViscosity(
                        dest=fluid, sources=all, c0=self.c0,
                        alpha=self.alpha
                    ))
            if self.nu > 0.0:
                eq4.append(
                    MomentumEquationViscosity(
                        dest=fluid, sources=all, nu=self.nu
                    ))
                if self.solids:
                    eq4.append(
                        SolidWallNoSlipBC(
                            dest=fluid, sources=self.solids, nu=self.nu
                        ))
            eq4.append(
                MomentumEquationArtificialStress(
                    dest=fluid, sources=self.fluids, dim=self.dim
                ))
        stage2.append(Group(equations=eq4, real=True))

        return MultiStageEquations([stage1, stage2])

    def setup_properties(self, particles, clean=True):
        particle_arrays = dict([(p.name, p) for p in particles])
        dummy = get_particle_array_gtvf(name='junk')
        props = list(dummy.properties.keys())
        props += [dict(name=p, stride=v) for p, v in dummy.stride.items()]
        output_props = dummy.output_property_arrays
        for fluid in self.fluids:
            pa = particle_arrays[fluid]
            self._ensure_properties(pa, props, clean)
            pa.set_output_arrays(output_props)

        solid_props = ['uf', 'vf', 'wf', 'vg', 'ug', 'wij', 'wg', 'V']
        props += solid_props
        for solid in self.solids:
            pa = particle_arrays[solid]
            self._ensure_properties(pa, props, clean)
            pa.set_output_arrays(output_props)
